#!/usr/bin/env python
#__init__.py
#
# Parallel Python Software: http://www.parallelpython.com
# Copyright (c) 2005-2009, Vitalii Vanovschi
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the author nor the names of its contributors
#      may be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
"""
Parallel Python Software, Execution Server.

http://www.parallelpython.com - updates, documentation, examples and support
forums
"""

from __future__ import absolute_import
from __future__ import print_function
# from six import text_type as str
from six.moves import range
# import six.moves.cPickle as pickle
import six.moves._thread
from io import TextIOWrapper

import os
import json
import logging
import inspect
import sys
import types
import time
import atexit
import traceback

# import user
from . import pptransport
from . import ppauto

copyright = 'Copyright (c) 2005-2009 Vitalii Vanovschi. All rights reserved'
version = '1.5.7'

# reconnect persistent rworkers in 5 sec
_RECONNECT_WAIT_TIME = 5

# we need to have set even in Python 2.3
try:
    set
except NameError:
    from sets import Set as set  # @UnresolvedImport

_USE_SUBPROCESS = False
try:
    import subprocess
    _USE_SUBPROCESS = True
except ImportError:
    import popen2


class _Task(object):
    """
    Class describing single task (job)
    """
    def __init__(self, server, tid, callback=None, callbackargs=(), group='default'):
        """
        Initializes the task.
        """
        self.lock = six.moves._thread.allocate_lock()
        self.lock.acquire()
        self.tid = tid
        self.server = server
        self.callback = callback
        self.callbackargs = callbackargs
        self.group = group
        self.finished = False
        self.unpickled = False
        self.worker_pid = -1
        self.cancelled = False

    def finalize(self, sresult):
        """
        Finalizes the task.

        For internal use only
        """
        if self.cancelled:
            self.sresult = None
        else:
            self.sresult = sresult
        if self.callback:
            self.__unpickle()
        self.lock.release()
        self.finished = True

    def __call__(self, raw_result=False):
        """
        Retrieves result of the task.
        """
        self.wait()

        if not self.unpickled and not raw_result:
            self.__unpickle()

        if raw_result:
            return self.sresult
        else:
            return self.result

    def wait(self):
        """
        Waits for the task.
        """
        if not self.finished:
            self.lock.acquire()
            self.lock.release()

    def __unpickle(self):
        """
        Unpickles the result of the task.
        """
        if self.sresult is not None:
            raw = self.sresult
            try:
                self.result, sout = json.loads(raw)['v']
            except:
                traceback.print_exc()
                print(raw)
            if len(sout) > 0:
                print(sout, end=' ')
        else:
            self.result = None
        self.unpickled = True
        if self.callback:
            args = self.callbackargs + (self.result, )
            self.callback(*args)


class _Worker(object):
    """
    Local worker class.
    """
    command = [sys.executable, '-u', 'bpworker.py']

    def __init__(self, restart_on_free, pickle_proto):
        """
        Initializes local worker.
        """
        self.restart_on_free = restart_on_free
        self.pickle_proto = pickle_proto
        self.start()

    def start(self):
        """
        Starts local worker.
        """
        if _USE_SUBPROCESS:
            if sys.platform.startswith('win'):
                import win32process  # @UnresolvedImport
                my_env = os.environ
                my_env['PYTHONIOENCODING'] = 'latin1'
                proc = subprocess.Popen(
                    self.command,
                    shell=False,
                    stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    universal_newlines=False,
                    creationflags=win32process.CREATE_NO_WINDOW,
                    env=my_env,
                )
                self.t = pptransport.PipeTransport(proc.stdout, proc.stdin)
            else:
                if six.PY2:
                    my_env = os.environ
                    my_env['PYTHONIOENCODING'] = 'latin1'
                    proc = subprocess.Popen(
                        self.command,
                        shell=False,
                        stdin=subprocess.PIPE,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        universal_newlines=False,
                        env=my_env,
                    )
                    self.t = pptransport.PipeTransport(proc.stdout, proc.stdin)
                else:
                    my_env = os.environ
                    my_env['PYTHONIOENCODING'] = 'latin1'
                    proc = subprocess.Popen(
                        self.command,
                        shell=False,
                        stdin=subprocess.PIPE,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        universal_newlines=False,
                        env=my_env,
                    )
                    self.t = pptransport.PipeTransport(proc.stdout, proc.stdin)
        else:
            self.t = pptransport.PipeTransport(*popen2.popen3(self.command)[:2])
        self.pid = int(self.t.receive())
        self.is_free = True
        logging.info('Started new worker: %s with %s', self.pid, self.t)

    def stop(self):
        """
        Stops local worker.
        """
        logging.info('Stop worker: %s with %s', self.pid, self.t)
        self.is_free = False
        self.t.close()

    def restart(self):
        """
        Restarts local worker.
        """
        self.stop()
        self.start()

    def free(self):
        """
        Frees local worker.
        """
        if self.restart_on_free:
            self.restart()
        else:
            self.is_free = True


class _RWorker(pptransport.CSocketTransport):
    """
    Remote worker class.
    """
    def __init__(self, host, port, secret, message=None, persistent=True):
        """
        Initializes remote worker.
        """
        self.persistent = persistent
        self.host = host
        self.port = port
        self.secret = secret
        self.address = (host, port)
        self.id = host + ':' + six.text_type(port)
        logging.info('Creating Rworker id=%s persistent=%s' % (self.id, persistent))
        self.connect(message)
        self.is_free = True

    def __del__(self):
        """
        Closes connection with remote server.
        """
        self.close()

    def connect(self, message=None):
        """
        Connects to a remote server.
        """
        while True:
            try:
                pptransport.SocketTransport.__init__(self)
                self._connect(self.host, self.port)
                if not self.authenticate(self.secret):
                    logging.error('Authentication failed for host=%s, port=%s' % (self.host, self.port))
                    return False
                if message:
                    self.send(message)
                self.is_free = True
                return True
            except:
                if not self.persistent:
                    logging.info('Deleting from queue Rworker %s' % (self.id, ))
                    return False
                logging.info('Failed to reconnect with '
                             '(host=%s, port=%i), will try again in %i s' % (self.host, self.port, _RECONNECT_WAIT_TIME), )
                time.sleep(_RECONNECT_WAIT_TIME)


class _Statistics(object):
    """
    Class to hold execution statisitcs for a single node.
    """
    def __init__(self, ncpus, rworker=None):
        """
        Initializes statistics for a node.
        """
        self.ncpus = ncpus
        self.time = 0.0
        self.njobs = 0
        self.rworker = rworker


class Template(object):
    """
    Template class.
    """
    def __init__(self, job_server, func, depfuncs=(), modules=(), callback=None, callbackargs=(), group='default', globals=None):
        """
        Creates Template instance.

        jobs_server - pp server for submitting jobs
        func - function to be executed
        depfuncs - tuple with functions which might be called from 'func'
        modules - tuple with module names to import
        callback - callback function which will be called with argument
                list equal to callbackargs+(result,)
                as soon as calculation is done
        callbackargs - additional arguments for callback function
        group - job group, is used when wait(group) is called to wait for
        jobs in a given group to finish
        globals - dictionary from which all modules, functions and classes
        will be imported, for instance: globals=globals()
        """
        self.job_server = job_server
        self.func = func
        self.depfuncs = depfuncs
        self.modules = modules
        self.callback = callback
        self.callbackargs = callbackargs
        self.group = group
        self.globals = globals

    def submit(self, *args):
        """
        Submits function with *arg arguments to the execution queue.
        """
        return self.job_server.submit(self.func, args, self.depfuncs, self.modules, self.callback, self.callbackargs, self.group, self.globals)


class Server(object):
    """
    Parallel Python SMP execution server class.
    """

    default_port = 60000
    default_secret = 'epo20pdosl;dksldkmm'

    def __init__(self, ncpus='autodetect', ppservers=(), secret=None, loglevel=logging.WARNING, logstream=sys.stderr, logfile=None, restart=False, proto=0):
        """
        Creates Server instance.

        ncpus - the number of worker processes to start on the local
                computer, if parameter is omitted it will be set to
                the number of processors in the system
        ppservers - list of active parallel python execution servers
                to connect with
        secret - passphrase for network connections, if omitted a default
                passphrase will be used. It's highly recommended to use a
                custom passphrase for all network connections.
        loglevel - logging level
        logstream - log stream destination
        logfile - if not None, will log to file instead of logstream
        restart - wheather to restart worker process after each task completion
        proto - protocol number for pickle module

        With ncpus = 1 all tasks are executed consequently
        For the best performance either use the default "autodetect" value
        or set ncpus to the total number of processors in the system
        """

        if not isinstance(ppservers, tuple):
            raise TypeError('ppservers argument must be a tuple')

        self.__initlog(loglevel, logstream=logstream, logfile=logfile)
        logging.info('Creating server instance (pp-' + version + ')' + repr(self))
        self.__tid = 0
        self.__active_tasks = 0
        self.__active_tasks_lock = six.moves._thread.allocate_lock()
        self.__queue = []
        self.__queue_lock = six.moves._thread.allocate_lock()
        self.__workers = []
        self.__rworkers = []
        self.__rworkers_reserved = []
        self.__rworkers_reserved4 = []
        self.__sourcesHM = {}
        self.__sfuncHM = {}
        self.__waittasks = []
        self.__waittasks_lock = six.moves._thread.allocate_lock()
        self.__exiting = False
        self.__accurate_stats = True
        self.autopp_list = {}
        self.__active_rworkers_list_lock = six.moves._thread.allocate_lock()
        self.__restart_on_free = restart
        self.__pickle_proto = proto

        # add local directory and sys.path to PYTHONPATH
        pythondirs = [os.getcwd()] + sys.path

        if 'PYTHONPATH' in os.environ and os.environ['PYTHONPATH']:
            pythondirs += os.environ['PYTHONPATH'].split(os.pathsep)
        os.environ['PYTHONPATH'] = os.pathsep.join(set(pythondirs))

        atexit.register(self.destroy)
        self.__stats = {'local': _Statistics(0)}
        self.set_ncpus(ncpus)

        self.ppservers = []
        self.auto_ppservers = []

        for ppserver in ppservers:
            ppserver = ppserver.split(':')
            host = ppserver[0]
            if len(ppserver) > 1:
                port = int(ppserver[1])
            else:
                port = Server.default_port
            if host.find('*') == -1:
                self.ppservers.append((host, port))
            else:
                if host == '*':
                    host = '*.*.*.*'
                interface = host.replace('*', '0')
                broadcast = host.replace('*', '255')
                self.auto_ppservers.append(((interface, port), (broadcast, port)))
        self.__stats_lock = six.moves._thread.allocate_lock()
        if secret is not None:
            # if not isinstance(secret, bytes):
            #     raise TypeError("secret must be of a string type")
            # self.secret = str(secret)
            self.secret = six.text_type(secret)
#         elif hasattr(user, "pp_secret"):
#             secret = user["pp_secret"]
#             if not isinstance(secret, bytes):
#                 raise TypeError("secret must be of a string type")
#             self.secret = str(secret)
        else:
            self.secret = Server.default_secret
        self.__connect()
        self.__creation_time = time.time()
        logging.info('pp local server started with %d workers' % (self.__ncpus, ))

    def submit(self, func, args=(), depfuncs=(), modules=(), callback=None, callbackargs=(), group='default', globals=None):
        """
        Submits function to the execution queue.

        func - function to be executed
        args - tuple with arguments of the 'func'
        depfuncs - tuple with functions which might be called from 'func'
        modules - tuple with module names to import
        callback - callback function which will be called with argument
                list equal to callbackargs+(result,)
                as soon as calculation is done
        callbackargs - additional arguments for callback function
        group - job group, is used when wait(group) is called to wait for
        jobs in a given group to finish
        globals - dictionary from which all modules, functions and classes
        will be imported, for instance: globals=globals()
        """
        # perform some checks for frequent mistakes
        if self.__exiting:
            raise RuntimeError('Cannot submit jobs: server'
                               ' instance has been destroyed', )

        if not isinstance(args, tuple):
            raise TypeError('args argument must be a tuple')

        if not isinstance(depfuncs, tuple):
            raise TypeError('depfuncs argument must be a tuple')

        if not isinstance(modules, tuple):
            raise TypeError('modules argument must be a tuple')

        if not isinstance(callbackargs, tuple):
            raise TypeError('callbackargs argument must be a tuple')

        for module in modules:
            if not isinstance(module, six.string_types):
                raise TypeError('modules argument must be a list of strings')

        tid = self.__gentid()

        if globals:
            modules += tuple(self.__find_modules('', globals))
            modules = tuple(set(modules))
            self.__logger.info('Task %i will autoimport next modules: %s' % (tid, six.text_type(modules)))
            for object1 in globals.values():
                if isinstance(object1, types.FunctionType) \
                        or isinstance(object1, type):
                    depfuncs += (object1, )

        task = _Task(self, tid, callback, callbackargs, group)

        self.__waittasks_lock.acquire()
        self.__waittasks.append(task)
        self.__waittasks_lock.release()

        # if the function is a method of a class add self to the arguments list
        if isinstance(func, types.MethodType) and func.__self__ is not None:
            args = (func.__self__, ) + args

        # if there is an instance of a user defined class in the arguments add
        # whole class to dependancies
        # for arg in args:
        #     # Checks for both classic or new class instances
        #     if isinstance(arg, object) or str(type(arg))[:6] == "<class":
        #         depfuncs += (arg.__class__, )

        # if there is a function in the arguments add this
        # function to dependancies
        for arg in args:
            if isinstance(arg, types.FunctionType):
                depfuncs += (arg, )

        sfunc = self.__dumpsfunc((func, ) + depfuncs, modules)
        # sargs = pickle.dumps(args, self.__pickle_proto)
        sargs = json.dumps({
            'v': args,
        })

        self.__queue_lock.acquire()
        self.__queue.append((task, sfunc, sargs))
        self.__queue_lock.release()

        self.__logger.info("Task %i submited, function='%s'" % (tid, func.__name__))
        self.__scheduler()
        return task

    def cancel(self, taskID):
        """
        Cancel a task.
        """

        wtask = None
        self.__waittasks_lock.acquire()
        for i in range(len(self.__waittasks)):
            wtask = self.__waittasks[i]
            if wtask.tid == taskID:
                self.__waittasks[i].cancelled = True
                break
        self.__waittasks_lock.release()

        if not wtask:
            return

        if wtask.worker_pid:
            for worker in self.__workers:
                if worker.pid == wtask.worker_pid:

                    nworker = _Worker(self.__restart_on_free, self.__pickle_proto)
                    self.__workers.append(nworker)
                    self.__logger.info('Started new worker, new process %i created' % (nworker.pid))

                    worker.t.exiting = True
                    if sys.platform.startswith('win'):
                        os.popen('TASKKILL /PID ' + six.text_type(worker.pid) + ' /F')
                    else:
                        try:
                            os.kill(worker.pid, 9)
                            os.waitpid(worker.pid, 0)
                        except:
                            pass

                    self.__workers.remove(worker)
                    self.__logger.info('Running task %i cancelled, process %i killed' % (taskID, wtask.worker_pid))
                    break

        else:

            self.__queue_lock.acquire()
            for i in range(len(self.__queue)):
                if self.__queue[i][0].tid == taskID:
                    self.__queue.pop(i)
            self.__queue_lock.release()

            self.__waittasks_lock.acquire()
            for i in range(len(self.__waittasks)):
                if self.__waittasks[i].tid == taskID:
                    self.__waittasks.pop(i)
                    break
            self.__waittasks_lock.release()

            self.__logger.info('Scheduled task %i cancelled' % taskID)

        self.__scheduler()

    def wait(self, group=None):
        """
        Waits for all jobs in a given group to finish.

        If group is omitted waits for all jobs to finish
        """
        while True:
            self.__waittasks_lock.acquire()
            for task in self.__waittasks:
                if not group or task.group == group:
                    self.__waittasks_lock.release()
                    task.wait()
                    break
            else:
                self.__waittasks_lock.release()
                break

    def get_ncpus(self):
        """
        Returns the number of local worker processes (ppworkers)
        """
        return self.__ncpus

    def set_ncpus(self, ncpus='autodetect'):
        """
        Sets the number of local worker processes (ppworkers)

        ncpus - the number of worker processes, if parammeter is omitted
                it will be set to the number of processors in the system
        """
        if ncpus == 'autodetect':
            ncpus = self.__detect_ncpus()
        if not isinstance(ncpus, int):
            raise TypeError("ncpus must have 'int' type")
        if ncpus < 0:
            raise ValueError('ncpus must be an integer > 0')
        if ncpus > len(self.__workers):
            try:
                self.__workers.extend([_Worker(self.__restart_on_free, self.__pickle_proto) for x in range(ncpus - len(self.__workers))])
            except:
                traceback.print_exc()
        self.__stats['local'].ncpus = ncpus
        self.__ncpus = ncpus

    def get_active_nodes(self):
        """
        Returns active nodes as a dictionary.

        [keys - nodes, values - ncpus]
        """
        active_nodes = {}
        for node, stat in self.__stats.items():
            if node == 'local' or node in self.autopp_list \
                    and self.autopp_list[node]:
                active_nodes[node] = stat.ncpus
        return active_nodes

    def get_stats(self):
        """
        Returns job execution statistics as a dictionary.
        """
        for node, stat in self.__stats.items():
            if stat.rworker:
                try:
                    stat.rworker.send('TIME')
                    stat.time = float(stat.rworker.receive())
                except:
                    self.__accurate_stats = False
                    stat.time = 0.0
        return self.__stats

    def print_stats(self):
        """
        Prints job execution statistics.

        Useful for benchmarking on clusters
        """

        print('Job execution statistics:')
        walltime = time.time() - self.__creation_time
        statistics = list(self.get_stats().items())
        totaljobs = 0.0
        for ppserver, stat in statistics:
            totaljobs += stat.njobs
        print(
            ' job count | % of all jobs | job time sum | ' \
            'time per job | job server',
        )
        for ppserver, stat in statistics:
            if stat.njobs:
                print(
                    '    %6i |        %6.2f |     %8.4f |  %11.6f | %s' \
                    % (
                        stat.njobs, 100.0 * stat.njobs / totaljobs, stat.time,
                        stat.time / stat.njobs, ppserver, ),
                )
        print('Time elapsed since server creation', walltime)

        if not self.__accurate_stats:
            print(
                'WARNING: statistics provided above is not accurate' \
                  ' due to job rescheduling',
            )
        print()

    # all methods below are for internal use only

    def insert(self, sfunc, sargs, task=None):
        """
        Inserts function into the execution queue.

        It's intended for internal use only (ppserver.py).
        """
        if not task:
            tid = self.__gentid()
            task = _Task(self, tid)
        self.__queue_lock.acquire()
        self.__queue.append((task, sfunc, sargs))
        self.__queue_lock.release()

        self.__logger.info('Task %i inserted' % (task.tid, ))
        self.__scheduler()
        return task

    def connect1(self, host, port, persistent=True):
        """
        Conects to a remote ppserver specified by host and port.
        """
        try:
            rworker = _RWorker(host, port, self.secret, 'STAT', persistent)
            ncpus = int(rworker.receive())
            hostid = host + ':' + six.text_type(port)
            self.__stats[hostid] = _Statistics(ncpus, rworker)

            for x in range(ncpus):
                rworker = _RWorker(host, port, self.secret, 'EXEC', persistent)
                self.__update_active_rworkers(rworker.id, 1)
                # append is atomic - no need to lock self.__rworkers
                self.__rworkers.append(rworker)
            # creating reserved rworkers
            for x in range(ncpus):
                rworker = _RWorker(host, port, self.secret, 'EXEC', persistent)
                self.__update_active_rworkers(rworker.id, 1)
                self.__rworkers_reserved.append(rworker)
            # creating reserved4 rworkers
            for x in range(ncpus*0):
                rworker = _RWorker(host, port, self.secret, 'EXEC', persistent)
                #                    self.__update_active_rworkers(rworker.id, 1)
                self.__rworkers_reserved4.append(rworker)
            logging.info('Connected to ppserver (host=%s, port=%i) \
                    with %i workers' % (host, port, ncpus))
            self.__scheduler()
        except:
            pass
#            sys.excepthook(*sys.exc_info())

    def __connect(self):
        """
        Connects to all remote ppservers.
        """
        for ppserver in self.ppservers:
            six.moves._thread.start_new_thread(self.connect1, ppserver)

        discover = ppauto.Discover(self, True)
        for ppserver in self.auto_ppservers:
            six.moves._thread.start_new_thread(discover.run, ppserver)

    def __detect_ncpus(self):
        """
        Detects the number of effective CPUs in the system.
        """
        # for Linux, Unix and MacOS
        if hasattr(os, 'sysconf'):
            if 'SC_NPROCESSORS_ONLN' in os.sysconf_names:
                #Linux and Unix
                ncpus = os.sysconf('SC_NPROCESSORS_ONLN')
                if isinstance(ncpus, int) and ncpus > 0:
                    return ncpus
            else:
                # MacOS X
                return int(os.popen2('sysctl -n hw.ncpu')[1].read())  # @UndefinedVariable
        # for Windows
        if 'NUMBER_OF_PROCESSORS' in os.environ:
            ncpus = int(os.environ['NUMBER_OF_PROCESSORS'])
            if ncpus > 0:
                return ncpus
        # return the default value
        return 1

    def __initlog(self, loglevel, logstream=None, logfile=None):
        """
        Initializes logging facility.
        """
        if logfile:
            log_handler = logging.FileHandler(logfile)
        else:
            log_handler = logging.StreamHandler(logstream)
        log_handler.setLevel(loglevel)
        LOG_FORMAT = '        %(levelname)s %(message)s'
        log_handler.setFormatter(logging.Formatter(LOG_FORMAT))
        self.__logger = logging.getLogger('')
        self.__logger.addHandler(log_handler)
        self.__logger.setLevel(loglevel)

    def __dumpsfunc(self, funcs, modules):
        """
        Serializes functions and modules.
        """
        hashs = hash(funcs + modules)
        if hashs not in self.__sfuncHM:
            sources = [self.__get_source(func) for func in funcs]
            # self.__sfuncHM[hashs] = pickle.dumps(
            #     (funcs[0].__name__, sources, modules),
            #     self.__pickle_proto)
            self.__sfuncHM[hashs] = json.dumps({
                'v': (funcs[0].__name__, sources, modules),
            })
        return self.__sfuncHM[hashs]

    def __find_modules(self, prefix, dict):
        """
        recursively finds all the modules in dict.
        """
        modules = []
        for name, object in dict.items():
            if isinstance(object, types.ModuleType) \
                    and name not in ('__builtins__', 'pp'):
                if object.__name__ == prefix + name or prefix == '':
                    modules.append(object.__name__)
                    modules.extend(self.__find_modules(object.__name__ + '.', object.__dict__))
        return modules

    def __scheduler(self):
        """
        Schedules jobs for execution.
        """
        self.__queue_lock.acquire()
        while self.__queue:
            if self.__active_tasks < self.__ncpus:
                # TODO: select a job number on the basis of heuristic
                task = self.__queue.pop(0)
                for worker in self.__workers:
                    if worker.is_free:
                        worker.is_free = False
                        break
                else:
                    self.__logger.error('There are no free workers left')
                    raise RuntimeError('Error: No free workers')
                self.__add_to_active_tasks(1)
                try:
                    self.__stats['local'].njobs += 1
                    six.moves._thread.start_new_thread(self.__run, task + (worker, ))
                except:
                    pass
                for i in range(len(self.__waittasks)):
                    if self.__waittasks[i].tid == task[0].tid:
                        self.__waittasks[i].worker_pid = worker.pid
                        break
            else:
                for rworker in self.__rworkers:
                    if rworker.is_free:
                        rworker.is_free = False
                        task = self.__queue.pop(0)
                        self.__stats[rworker.id].njobs += 1
                        six.moves._thread.start_new_thread(self.__rrun, task + (rworker, ))
                        break
                else:
                    if len(self.__queue) > self.__ncpus:
                        for rworker in self.__rworkers_reserved:
                            if rworker.is_free:
                                rworker.is_free = False
                                task = self.__queue.pop(0)
                                self.__stats[rworker.id].njobs += 1
                                six.moves._thread.start_new_thread(self.__rrun, task + (rworker, ))
                                break
                        else:
                            break
                            # this code will not be executed
                            # and is left for further releases
                            if len(self.__queue) > self.__ncpus*0:
                                for rworker in self.__rworkers_reserved4:
                                    if rworker.is_free:
                                        rworker.is_free = False
                                        task = self.__queue.pop(0)
                                        self.__stats[rworker.id].njobs += 1
                                        six.moves._thread.start_new_thread(self.__rrun, task + (rworker, ))
                                        break
                    else:
                        break

        self.__queue_lock.release()

    def __get_source(self, func):
        """
        Fetches source of the function.
        """
        hashf = hash(func)
        if hashf not in self.__sourcesHM:
            # get lines of the source and adjust indent
            sourcelines = inspect.getsourcelines(func)[0]
            # remove indentation from the first line
            sourcelines[0] = sourcelines[0].lstrip()
            self.__sourcesHM[hashf] = ''.join(sourcelines)
        return self.__sourcesHM[hashf]

    def __rrun(self, job, sfunc, sargs, rworker):
        """
        Runs a job remotelly.
        """
        self.__logger.info('Task (remote) %i started' % (job.tid, ))

        try:
            rworker.send(sfunc)
            rworker.send(sargs)
            sresult = rworker.receive()
            rworker.is_free = True
        except:
            self.__logger.info('Task %i failed due to broken network '
                               'connection - rescheduling' % (job.tid, ), )
            self.insert(sfunc, sargs, job)
            self.__scheduler()
            self.__update_active_rworkers(rworker.id, -1)
            if rworker.connect('EXEC'):
                self.__update_active_rworkers(rworker.id, 1)
                self.__scheduler()
            return

        job.finalize(sresult)

        # remove the job from the waiting list
        if self.__waittasks:
            self.__waittasks_lock.acquire()
            self.__waittasks.remove(job)
            self.__waittasks_lock.release()

        self.__logger.info('Task (remote) %i ended' % (job.tid, ))
        self.__scheduler()

    def __run(self, job, sfunc, sargs, worker):
        """
        Runs a job locally.
        """

        if self.__exiting:
            return
        self.__logger.info('Task %i started' % (job.tid, ))

        start_time = time.time()

        sresult = None
        try:
            worker.t.send(sfunc)
            worker.t.send(sargs)
            sresult = worker.t.receive()
        except:
            if self.__exiting:
                return
            else:
                if not job.cancelled:
                    sys.excepthook(*sys.exc_info())

        worker.free()

        job.finalize(sresult)

        # remove the job from the waiting list
        if self.__waittasks:
            self.__waittasks_lock.acquire()
            self.__waittasks.remove(job)
            self.__waittasks_lock.release()

        self.__add_to_active_tasks(-1)
        if not self.__exiting:
            self.__stat_add_time('local', time.time() - start_time)
        self.__logger.info('Task %i ended' % (job.tid, ))
        self.__scheduler()

    def __add_to_active_tasks(self, num):
        """
        Updates the number of active tasks.
        """
        self.__active_tasks_lock.acquire()
        self.__active_tasks += num
        self.__active_tasks_lock.release()

    def __stat_add_time(self, node, time_add):
        """
        Updates total runtime on the node.
        """
        self.__stats_lock.acquire()
        self.__stats[node].time += time_add
        self.__stats_lock.release()

    def __stat_add_job(self, node):
        """
        Increments job count on the node.
        """
        self.__stats_lock.acquire()
        self.__stats[node].njobs += 1
        self.__stats_lock.release()

    def __update_active_rworkers(self, id, count):
        """
        Updates list of active rworkers.
        """
        self.__active_rworkers_list_lock.acquire()

        if id not in self.autopp_list:
            self.autopp_list[id] = 0
        self.autopp_list[id] += count

        self.__active_rworkers_list_lock.release()

    def __gentid(self):
        """
        Generates a unique job ID number.
        """
        self.__tid += 1
        return self.__tid - 1

    def destroy(self):
        """
        Kills ppworkers and closes open files.
        """
        self.__exiting = True
        self.__queue_lock.acquire()
        self.__queue = []
        self.__queue_lock.release()

        for worker in self.__workers:
            worker.t.exiting = True
            if sys.platform.startswith('win'):
                os.popen('TASKKILL /PID ' + six.text_type(worker.pid) + ' /F')
            else:
                try:
                    os.kill(worker.pid, 9)
                    os.waitpid(worker.pid, 0)
                except:
                    pass


# Parallel Python Software: http://www.parallelpython.com



#!/usr/bin/env python
#__init__.py
#
#
# Copyright (C) 2007-2008 Francois Aucamp, Meraka Institute, CSIR
# See AUTHORS for all authors and contact information.
#
# License: GNU Lesser General Public License, version 3 or later; see COPYING
#          included in this archive for details.
#
# This library is free software, distributed under the terms of
# the GNU Lesser General Public License Version 3, or any later version.
# See the COPYING file included in this archive
#
# The docstrings in this module contain epytext markup; API documentation
# may be created by processing this file with epydoc: http://epydoc.sf.net
"""
Kademlia DHT implementation.

This package contains Entangled's implementation of the Kademlia
distributed hash table (DHT).

The main modules in this package are "C{node}" (which contains the Kademlia
implementation's main interface, namely the C{Node} class), "C{datastore}"
(physical data storage mechanisms), "C{constants}" (several constant values
defining the Kademlia network), "C{routingtable}" (different Kademlia routing
table implementations) and "C{protocol}" (actual network communications).

The Node class is directly exposed in the main Entangled package
("C{entangled}") as KademliaNode, and as Node in this package
("C{entangled.kademlia}"). It is designed to be customizable; the data storage
mechansims may (and should) be directly specified by client applications via
the node's contructor. The same holds true for the node's routing table and
network protocol used. This potentially allows the Kademlia node to be used
with a TCP-based protocol, instead of the provided UDP-based one.

Client applications should also modify the values found in
C{entangled.kademlia.constants} to suit their needs. Refer to the C{constants}
module for documentation on what these values control.
"""

from __future__ import absolute_import
from .node import Node  # @UnresolvedImport
from .datastore import DictDataStore, SQLiteVersionedJsonDataStore  # @UnresolvedImport



import os
import time
import threading
import socks
import sqlite3
import traceback
import logging
import json

#------------------------------------------------------------------------------

from twisted.internet.defer import Deferred
from twisted.internet import reactor

#------------------------------------------------------------------------------

from bitdust_forks.Bismuth import node as bismuth_node  # @UnresolvedImport
from bitdust_forks.Bismuth import connectionmanager  # @UnresolvedImport
from bitdust_forks.Bismuth import mempool as mp  # @UnresolvedImport
from bitdust_forks.Bismuth import apihandler  # @UnresolvedImport
from bitdust_forks.Bismuth import dbhandler  # @UnresolvedImport
from bitdust_forks.Bismuth import connections  # @UnresolvedImport
from bitdust_forks.Bismuth import options  # @UnresolvedImport
from bitdust_forks.Bismuth import peershandler  # @UnresolvedImport
from bitdust_forks.Bismuth import plugins  # @UnresolvedImport
from bitdust_forks.Bismuth.libs import node as _node  # @UnresolvedImport
from bitdust_forks.Bismuth.libs import logger  # @UnresolvedImport
from bitdust_forks.Bismuth.libs import keys  # @UnresolvedImport
from bitdust_forks.Bismuth.modules import config as modules_config  # @UnresolvedImport
from bitdust_forks.Bismuth.bismuthclient import rpcconnections  # @UnresolvedImport

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.main import settings
from bitdust.main import config as main_conf

from bitdust.blockchain import known_bismuth_nodes

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 12

#------------------------------------------------------------------------------

VERSION = '1.0.0.0'

_DataDirPath = None

#------------------------------------------------------------------------------


def init():
    global _DataDirPath
    _DataDirPath = settings.ServiceDir('bismuth_blockchain')
    create_config_file(_DataDirPath)
    starting_defer = Deferred()
    node_thread = threading.Thread(target=run, args=(_DataDirPath, starting_defer))
    node_thread.start()
    return starting_defer


def shutdown():
    global _DataDirPath

    config_path = os.path.join(_DataDirPath, 'config')
    custom_config_path = os.path.join(_DataDirPath, 'config_custom')

    config = options.Get()
    config.read(filename=config_path, custom_filename=custom_config_path)

    s = socks.socksocket()
    port = config.port

    count = 0
    while count < 1:
        try:
            s.connect(('127.0.0.1', int(port)))
            # print('Sending stop command...')
            connections.send(s, 'stop')
            # print('Stop command delivered.')
            break
        except Exception as e:
            lg.exc()
            # print('Cannot reach node', e)
            # time.sleep(0.1)
            count += 1

    s.close()
    return True


#------------------------------------------------------------------------------


def nod():
    return bismuth_node.node


#------------------------------------------------------------------------------


def run(data_dir_path, starting_defer):
    global _DataDirPath

    rpcconnections.LTIMEOUT = 20

    _DataDirPath = data_dir_path
    if not os.path.exists(data_dir_path):
        os.makedirs(data_dir_path)

    config_path = os.path.join(data_dir_path, 'config')
    custom_config_path = os.path.join(data_dir_path, 'config_custom')

    bismuth_node.node = _node.Node()
    node = bismuth_node.node
    bismuth_node.bootstrap = bootstrap

    node.app_version = VERSION

    node.FOUNDATION_MINERS = known_bismuth_nodes.foundation_miners()

    options.Get.defaults['heavy3_path'] = os.path.join(data_dir_path, 'heavy3a.bin')
    options.Get.defaults['mempool_path'] = os.path.join(data_dir_path, 'mempool.db')
    modules_config.Get.defaults['db_path'] = data_dir_path
    modules_config.Get.defaults['mempool_path'] = os.path.join(data_dir_path, 'mempool.db')

    node.data_dir_path = data_dir_path

    node.keys = keys.Keys()

    node.is_testnet = False
    node.is_regnet = False
    node.is_mainnet = True

    config = options.Get()
    config.read(filename=config_path, custom_filename=custom_config_path)

    node.version = config.version
    node.debug_level = config.debug_level

    node.logger = logger.Logger()
    if _Debug:
        # node.logger.app_log = custom_log(level_input=lg.get_loging_level(max(0, lg.get_debug_level() - 6), return_name=True))
        node.logger.app_log = custom_log(level_input=lg.get_loging_level(_DebugLevel - 6, return_name=True))
        # node.logger.app_log = custom_log(level_input='DEBUG')
    else:
        node.logger.app_log = custom_log(level_input='CRITICAL')
    node.logger.app_log.critical(f'Python version: {node.py_version}')

    node.host = config.node_ip
    node.port = config.port
    node.verify = config.verify
    node.thread_limit = config.thread_limit
    node.rebuild_db = config.rebuild_db
    node.debug = config.debug
    node.debug_level = config.debug_level
    node.pause = config.pause
    node.ledger_path = config.ledger_path
    node.hyper_path = config.hyper_path
    node.hyper_recompress = config.hyper_recompress
    node.tor = config.tor
    node.ram = config.ram
    node.version_allow = config.version_allow
    node.reveal_address = config.reveal_address
    node.terminal_output = config.terminal_output
    node.egress = config.egress
    node.genesis = config.genesis
    node.accept_peers = config.accept_peers
    node.full_ledger = config.full_ledger
    node.trace_db_calls = config.trace_db_calls
    node.old_sqlite = config.old_sqlite
    node.heavy = config.heavy
    node.heavy3_path = config.heavy3_path

    # node.logger.app_log.warning('Configuration settings loaded')

    # if not node.full_ledger and os.path.exists(node.ledger_path) and node.is_mainnet:
    #     os.remove(node.ledger_path)
    #     node.logger.app_log.warning('Removed full ledger for hyperblock mode')
    # if not node.full_ledger:
    #     node.logger.app_log.warning('Cloning hyperblocks to ledger file')
    #     shutil.copy(node.hyper_path, node.ledger_path)

    try:
        node.plugin_manager = plugins.PluginManager(app_log=node.logger.app_log, config=config, init=True)
        extra_commands = {}
        extra_commands = node.plugin_manager.execute_filter_hook('extra_commands_prefixes', extra_commands)

        setup_net_type(bismuth_node.node, data_dir_path, node.logger.app_log)

        bismuth_node.load_keys(data_dir=data_dir_path, wallet_filename=os.path.join(data_dir_path, 'node_key.json'))

        # node.logger.app_log.warning(f'Checking Heavy3 file, can take up to 5 minutes...')
        t_now = time.time()

        from bitdust_forks.Bismuth import mining_heavy3  # @UnresolvedImport
        # from bitdust_forks.Bismuth import digest  # @UnresolvedImport

        mining_heavy3.mining_open(node.heavy3_path)
        # digest.mining_heavy3.MMAP = mining_heavy3.MMAP
        # digest.mining_heavy3.RND_LEN = mining_heavy3.RND_LEN
        node.logger.app_log.warning(f'Heavy3 file is OK, loaded in %s seconds' % (time.time() - t_now))

        # node.logger.app_log.warning(f'Status: Starting node version {VERSION}')
        node.startup_time = time.time()
        try:

            node.peers = peershandler.Peers(node.logger.app_log, config=config, node=bismuth_node.node)
            node.peers.peerfile = bismuth_node.node.peerfile  # @UndefinedVariable
            node.peers.suggested_peerfile = bismuth_node.node.peerfile_suggested  # @UndefinedVariable

            node.apihandler = apihandler.ApiHandler(node.logger.app_log, config)
            mp.MEMPOOL = mp.Mempool(node.logger.app_log, config, bismuth_node.node.db_lock, bismuth_node.node.is_testnet, trace_db_calls=bismuth_node.node.trace_db_calls)
            bismuth_node.mp.MEMPOOL = mp.MEMPOOL

            check_db_for_bootstrap(bismuth_node.node)

            db_handler_initial = dbhandler.DbHandler(
                bismuth_node.node.index_db, bismuth_node.node.ledger_path, bismuth_node.node.hyper_path, bismuth_node.node.ram, bismuth_node.node.ledger_ram_file, node.logger, trace_db_calls=bismuth_node.node.trace_db_calls
            )
            bismuth_node.db_handler_initial = db_handler_initial

            try:
                bismuth_node.ledger_check_heights(bismuth_node.node, db_handler_initial)
            except:
                traceback.print_exc()

            if node.recompress:
                #todo: do not close database and move files, swap tables instead
                db_handler_initial.close()
                bismuth_node.recompress_ledger(bismuth_node.node)
                db_handler_initial = dbhandler.DbHandler(
                    bismuth_node.node.index_db, bismuth_node.node.ledger_path, bismuth_node.node.hyper_path, bismuth_node.node.ram, bismuth_node.node.ledger_ram_file, node.logger, trace_db_calls=bismuth_node.node.trace_db_calls
                )
                bismuth_node.db_handler_initial = db_handler_initial

            bismuth_node.ram_init(db_handler_initial)
            bismuth_node.node_block_init(db_handler_initial)
            bismuth_node.initial_db_check()

            if not node.is_regnet:
                bismuth_node.sequencing_check(db_handler_initial)

            if node.verify:
                bismuth_node.verify(db_handler_initial)

            bismuth_node.add_indices(db_handler_initial)

            db_handler_initial.close()

            if not node.tor:
                host = '0.0.0.0'
                port = int(node.port)

                bismuth_node.ThreadedTCPServer.allow_reuse_address = True
                bismuth_node.ThreadedTCPServer.daemon_threads = True
                bismuth_node.ThreadedTCPServer.timeout = 60
                bismuth_node.ThreadedTCPServer.request_queue_size = 100

                server = bismuth_node.ThreadedTCPServer((host, port), bismuth_node.ThreadedTCPRequestHandler)
                ip, node.port = server.server_address

                server_thread = threading.Thread(target=server.serve_forever)
                server_thread.daemon = True
                server_thread.start()

                node.logger.app_log.warning(f'Status: Server loop running on {ip}:{node.port}')

            else:
                node.logger.app_log.warning('Status: Not starting a local server to conceal identity on Tor network')

            connection_manager = connectionmanager.ConnectionManager(bismuth_node.node, mp)
            connection_manager.start()

        except Exception as e:
            node.logger.app_log.info(e)
            reactor.callFromThread(starting_defer.errback, e)  # @UndefinedVariable
            raise

    except Exception as e:
        node.logger.app_log.info(e)
        reactor.callFromThread(starting_defer.errback, e)  # @UndefinedVariable
        raise

    # node.logger.app_log.warning('Status: Bismuth loop running.')

    reactor.callFromThread(starting_defer.callback, True)  # @UndefinedVariable

    while True:
        if node.IS_STOPPING:
            if node.db_lock.locked():
                time.sleep(0.5)
            else:
                mining_heavy3.mining_close()
                node.logger.app_log.warning('Status: Securely disconnected main processes, subprocess termination in progress.')
                break
        time.sleep(0.1)

    try:
        bismuth_node.db_handler_initial.close()
    except:
        pass

    node.logger.app_log.warning('Status: Clean Stop')


def create_config_file(data_dir_path):
    config_path = os.path.join(data_dir_path, 'config')
    node_host = main_conf.conf().getString('services/bismuth-node/host', '127.0.0.1')
    node_port = main_conf.conf().getInt('services/bismuth-node/tcp-port', 15658)
    if _Debug:
        lg.args(_DebugLevel, node_host=node_host, node_port=node_port)
    config_src = '''debug=False
port={port}
verify=False
version=mainnet0001
version_allow=mainnet0001
thread_limit=64
rebuild_db=True
debug_level=DEBUG
purge=True
pause=5
hyper_path={hyper_path}
hyper_recompress=True
full_ledger=True
ledger_path={ledger_path}
ban_threshold=30
tor=False
allowed=127.0.0.1,192.168.0.1,any
ram=False
heavy=False
node_ip={node_ip}
light_ip={light_ip}
reveal_address=True
accept_peers=True
banlist=
whitelist=
nodes_ban_reset=5
terminal_output=False
gui_scaling=adapt
mempool_ram=False
egress=True
trace_db_calls={trace_db_calls}
heavy3_path={heavy3_path}'''.format(
        port=node_port,
        node_ip=node_host,
        hyper_path=os.path.join(data_dir_path, 'hyper.db'),
        ledger_path=os.path.join(data_dir_path, 'ledger.db'),
        heavy3_path=os.path.join(data_dir_path, 'heavy3a.bin'),
        light_ip='{"%s": "%d"}' % (node_host, node_port),
        trace_db_calls='True' if (_Debug and _DebugLevel >= 16) else 'False',
    )
    fout = open(config_path, 'w')
    fout.write(config_src)
    fout.flush()
    fout.close()


def setup_net_type(node, data_dir_path, app_log):
    """
    Adjust globals depending on mainnet, testnet or regnet
    """
    node.is_mainnet = True
    node.is_testnet = False
    node.is_regnet = False

    if 'testnet' in node.version or node.is_testnet:
        node.is_testnet = True
        node.is_mainnet = False
        node.version_allow = 'testnet'

    if 'regnet' in node.version or node.is_regnet:
        node.is_regnet = True
        node.is_testnet = False
        node.is_mainnet = False

    node.peerfile = os.path.join(data_dir_path, 'peers.json')
    node.peerfile_suggested = os.path.join(data_dir_path, 'suggested_peers.json')
    node.ledger_ram_file = 'file:ledger?mode=memory&cache=shared'
    node.index_db = os.path.join(data_dir_path, 'index.db')
    peerfile_data = known_bismuth_nodes.nodes_by_host().copy()
    app_log.warning(f'node host: {node.host} peerfile_data: {peerfile_data}')
    peerfile_data.pop(node.host, None)
    open(node.peerfile_suggested, 'w').write(json.dumps(peerfile_data))
    open(node.peerfile, 'w').write(json.dumps({'127.0.0.1': node.port}))
    # open(node.peerfile, 'w').write(json.dumps({}))


def bootstrap():
    global _DataDirPath

    try:
        hyper_path = os.path.join(_DataDirPath, 'hyper.db')
        ledger_path = os.path.join(_DataDirPath, 'ledger.db')
        index_path = os.path.join(_DataDirPath, 'index.db')

        INITIAL_DIFFICULTY = 10
        INITIAL_HYPER_DIFFICULTY = 10

        hdd = sqlite3.connect(ledger_path, timeout=1)
        hdd.text_factory = str
        hdd.execute('PRAGMA case_sensitive_like = 1;')
        hdd_cursor = hdd.cursor()
        hdd_cursor.execute('CREATE TABLE IF NOT EXISTS "misc" ("block_height" INTEGER, "difficulty" TEXT)')
        hdd_cursor.execute(
            'CREATE TABLE IF NOT EXISTS "transactions" ("block_height" INTEGER, "timestamp" NUMERIC, "address" TEXT, "recipient" TEXT, "amount" NUMERIC, "signature" TEXT, "public_key" TEXT, "block_hash" TEXT, "fee" NUMERIC, "reward" NUMERIC, "operation" TEXT, "openfield" TEXT)',
        )
        hdd_cursor.execute('CREATE INDEX "Timestamp Index" ON "transactions" ("timestamp")')
        hdd_cursor.execute('CREATE INDEX "Signature Index" ON "transactions" ("signature")')
        hdd_cursor.execute('CREATE INDEX "Reward Index" ON "transactions" ("reward")')
        hdd_cursor.execute('CREATE INDEX "Recipient Index" ON "transactions" ("recipient")')
        hdd_cursor.execute('CREATE INDEX "Openfield Index" ON "transactions" ("openfield")')
        hdd_cursor.execute('CREATE INDEX "Fee Index" ON "transactions" ("fee")')
        hdd_cursor.execute('CREATE INDEX "Block Height Index" ON "transactions" ("block_height")')
        hdd_cursor.execute('CREATE INDEX "Block Hash Index" ON "transactions" ("block_hash")')
        hdd_cursor.execute('CREATE INDEX "Amount Index" ON "transactions" ("amount")')
        hdd_cursor.execute('CREATE INDEX "Address Index" ON "transactions" ("address")')
        hdd_cursor.execute('CREATE INDEX "Operation Index" ON "transactions" ("operation")')
        hdd_cursor.execute('CREATE INDEX TXID4_Index ON transactions(substr(signature,1,4))')
        hdd_cursor.execute('CREATE INDEX "Misc Block Height Index" on misc(block_height)')
        hdd_cursor.execute('INSERT INTO misc (difficulty, block_height) VALUES ({},1)'.format(INITIAL_DIFFICULTY))
        hdd_cursor.execute(
            'INSERT INTO transactions VALUES (?,?,?,?,?,?,?,?,?,?,?,?)', (
                '1',
                '1683400177.01208',
                'genesis',
                options.GENESIS_ADDRESS,
                '0',
                'c5QPId20zONTTyIYKB7zIiCu+FqrtI9dXs/NFS7cCzW56Khd+6s1+TKu/54vEREnt35yl7AhgRPK8gCuQkN6iwKHXMPZTUvt7FLS28558Y1Au3Tlo2tIj394L/Zu+jQcRxI85QOH+u00kcPaDSmkj/38btUPvxW1o/BibZ3mfNCVkVVwLdoAgNdMXGStBnwmFrvf9odWEDT/bIK7cxmAXJgvXfKtmgQFMCBglGByRE51ZkGKrFMvzY42JY/MvT5q7mAvbAcENWe7KbSjLgOVHCaZf2FIfJRLGfncjKI+Z/jPasvzelAtFTtHK39aLO4KGqC36kr5/VAGqfGDOYZCMOwdYu9KNOmTfVaV7LWyX7I5DYIV1QOTd8+20gEN5AdDulbw9Bp2o/v6dJpcC4HceZNiWtaVovH5bHAWrgY9piNCtirkV/mwqiBnprzYYz34PVbm4KkPD3d9tnwWEz/eGZnMm5AjLx/fhetaCjyfZUPiXVQ8eP+mE0ukIjOrvNemB+MP/FqYDa5SzyJI34WlWu6S/BOCn0D0YlkWLywk48lbBs1J/guHQu7DxH+0PPTD0gi6jnPLio8Ch5A78riSVgLV1xhUxMSzJ7AFKZy2V17r2L1D7qDImTprCBGqihze9hRGXgZNCS0XiT4AGuVhnRNFzsC/mTUlaBHxfbig6tY=',
                'LS0tLS1CRUdJTiBQVUJMSUMgS0VZLS0tLS0KTUlJQ0lqQU5CZ2txaGtpRzl3MEJBUUVGQUFPQ0FnOEFNSUlDQ2dLQ0FnRUFuNmVjMVBmZjVWbDJ0QlJoY3RtSgp5TVVndG92QUQzcnBNVjcwWlp1Ukk2N0llUGJiYkoveVBnSVU4S0h4RkNEZ0h1eDJOQy9CU1VkNWpyTENQZzE2CktIcVZkbHR2SUwxcWduTmNpWTJOS0VoVkxRMGppQXpBTDgvWE9NNGhMUUtjbnk2dFRpUGVjMHk1ZG5BY2UxekoKNzhCVG1FRDJLVFVQdmJLblFJMS81NlNpU2QrU1Evcnp5aStlK2MwWllFaExTMk50TzVUd0oyV3IrNlY0VW5YcgpjT0U0UysrLzR3THIyTW5GR3d6dWNlWWVOa0NIT2N5ODZmbmFpQmVsbHY1YUN0Y3V4bHFjalFQdzFDU29MaUlICkN5MnhKYkZSeHJYU21qSlRxNG1WWFlmNlFKa0s0WDZPNEtpT2NZMDNRd2tuK1dkczk0c1M4Ny9QZ3NwZEFsSzYKcEF6MU9NZFlvSWJxUUlNR3NOaS9CSlh6VHpTYTF5UnpZZnB6LzlyOUxSYm5RS3hUSzlVSWlrYlY0Ri9JdU80NApBQ29SN3BWY29FQVlJYjRUcVFDUkFYS0VRQ2trRm9RZmNZaVBaOURTbjRiMjNtSEhMWENCWUJpek5wQVdxcXZxCnlSbzVxMVhRM3lKVXVlaVNEUnVZNldEa0MyaWlFWTdYTHlzWnErYnAxS1JjSENjNDBGMnNQU2RrWWozTGJZTUgKaWJsOFd0Q21NY0lFekVKOWdwWmU4RFNrV0tTVkNjNVg3YWpoODVFVU9zaG9KTWdSMGdaQ1FkRGFkaFh5S2xKQQpvZDcvV0JFWVhMa3ZSNWl1Tkk5aHl4Y3VYWkdEaWhNeUxLMmtXaWlIRDN0M3Zadjg4MjlJbDc5NURwTm9rNmhaCnUyS0F4UmtHR1U4aFBjR1hvVVF4eEdFQ0F3RUFBUT09Ci0tLS0tRU5EIFBVQkxJQyBLRVktLS0tLQ==',
                'fbc6d3d091d7c5ed745f499c548c103f03b96354c7bb3b3e982a128c',
                0,
                1,
                1,
                'genesis',
            )
        )
        hdd.commit()
        hdd.close()

        hyp = sqlite3.connect(hyper_path, timeout=1)
        hyp.text_factory = str
        hyp.execute('PRAGMA case_sensitive_like = 1;')
        hyp_cursor = hyp.cursor()
        hyp_cursor.execute('CREATE TABLE IF NOT EXISTS "misc" ("block_height" INTEGER, "difficulty" TEXT)')
        hyp_cursor.execute(
            'CREATE TABLE IF NOT EXISTS "transactions" ("block_height" INTEGER, "timestamp" NUMERIC, "address" TEXT, "recipient" TEXT, "amount" NUMERIC, "signature" TEXT, "public_key" TEXT, "block_hash" TEXT, "fee" NUMERIC, "reward" NUMERIC, "operation" TEXT, "openfield" TEXT)',
        )
        hyp_cursor.execute('CREATE INDEX "Timestamp Index" ON "transactions" ("timestamp")')
        hyp_cursor.execute('CREATE INDEX "Signature Index" ON "transactions" ("signature")')
        hyp_cursor.execute('CREATE INDEX "Reward Index" ON "transactions" ("reward")')
        hyp_cursor.execute('CREATE INDEX "Recipient Index" ON "transactions" ("recipient")')
        hyp_cursor.execute('CREATE INDEX "Openfield Index" ON "transactions" ("openfield")')
        hyp_cursor.execute('CREATE INDEX "Fee Index" ON "transactions" ("fee")')
        hyp_cursor.execute('CREATE INDEX "Block Height Index" ON "transactions" ("block_height")')
        hyp_cursor.execute('CREATE INDEX "Block Hash Index" ON "transactions" ("block_hash")')
        hyp_cursor.execute('CREATE INDEX "Amount Index" ON "transactions" ("amount")')
        hyp_cursor.execute('CREATE INDEX "Address Index" ON "transactions" ("address")')
        hyp_cursor.execute('CREATE INDEX "Operation Index" ON "transactions" ("operation")')
        hyp_cursor.execute('CREATE INDEX TXID4_Index ON transactions(substr(signature,1,4))')
        hyp_cursor.execute('CREATE INDEX "Misc Block Height Index" on misc(block_height)')
        hyp_cursor.execute('INSERT INTO misc (difficulty, block_height) VALUES ({},1)'.format(INITIAL_HYPER_DIFFICULTY))
        hyp_cursor.execute(
            'INSERT INTO transactions VALUES (?,?,?,?,?,?,?,?,?,?,?,?)', (
                '1',
                '1683400177.0120773',
                'genesis',
                options.GENESIS_ADDRESS,
                '0',
                'c5QPId20zONTTyIYKB7zIiCu+FqrtI9dXs/NFS7cCzW56Khd+6s1+TKu/54vEREnt35yl7AhgRPK8gCuQkN6iwKHXMPZTUvt7FLS28558Y1Au3Tlo2tIj394L/Zu+jQcRxI85QOH+u00kcPaDSmkj/38btUPvxW1o/BibZ3mfNCVkVVwLdoAgNdMXGStBnwmFrvf9odWEDT/bIK7cxmAXJgvXfKtmgQFMCBglGByRE51ZkGKrFMvzY42JY/MvT5q7mAvbAcENWe7KbSjLgOVHCaZf2FIfJRLGfncjKI+Z/jPasvzelAtFTtHK39aLO4KGqC36kr5/VAGqfGDOYZCMOwdYu9KNOmTfVaV7LWyX7I5DYIV1QOTd8+20gEN5AdDulbw9Bp2o/v6dJpcC4HceZNiWtaVovH5bHAWrgY9piNCtirkV/mwqiBnprzYYz34PVbm4KkPD3d9tnwWEz/eGZnMm5AjLx/fhetaCjyfZUPiXVQ8eP+mE0ukIjOrvNemB+MP/FqYDa5SzyJI34WlWu6S/BOCn0D0YlkWLywk48lbBs1J/guHQu7DxH+0PPTD0gi6jnPLio8Ch5A78riSVgLV1xhUxMSzJ7AFKZy2V17r2L1D7qDImTprCBGqihze9hRGXgZNCS0XiT4AGuVhnRNFzsC/mTUlaBHxfbig6tY=',
                'LS0tLS1CRUdJTiBQVUJMSUMgS0VZLS0tLS0KTUlJQ0lqQU5CZ2txaGtpRzl3MEJBUUVGQUFPQ0FnOEFNSUlDQ2dLQ0FnRUFuNmVjMVBmZjVWbDJ0QlJoY3RtSgp5TVVndG92QUQzcnBNVjcwWlp1Ukk2N0llUGJiYkoveVBnSVU4S0h4RkNEZ0h1eDJOQy9CU1VkNWpyTENQZzE2CktIcVZkbHR2SUwxcWduTmNpWTJOS0VoVkxRMGppQXpBTDgvWE9NNGhMUUtjbnk2dFRpUGVjMHk1ZG5BY2UxekoKNzhCVG1FRDJLVFVQdmJLblFJMS81NlNpU2QrU1Evcnp5aStlK2MwWllFaExTMk50TzVUd0oyV3IrNlY0VW5YcgpjT0U0UysrLzR3THIyTW5GR3d6dWNlWWVOa0NIT2N5ODZmbmFpQmVsbHY1YUN0Y3V4bHFjalFQdzFDU29MaUlICkN5MnhKYkZSeHJYU21qSlRxNG1WWFlmNlFKa0s0WDZPNEtpT2NZMDNRd2tuK1dkczk0c1M4Ny9QZ3NwZEFsSzYKcEF6MU9NZFlvSWJxUUlNR3NOaS9CSlh6VHpTYTF5UnpZZnB6LzlyOUxSYm5RS3hUSzlVSWlrYlY0Ri9JdU80NApBQ29SN3BWY29FQVlJYjRUcVFDUkFYS0VRQ2trRm9RZmNZaVBaOURTbjRiMjNtSEhMWENCWUJpek5wQVdxcXZxCnlSbzVxMVhRM3lKVXVlaVNEUnVZNldEa0MyaWlFWTdYTHlzWnErYnAxS1JjSENjNDBGMnNQU2RrWWozTGJZTUgKaWJsOFd0Q21NY0lFekVKOWdwWmU4RFNrV0tTVkNjNVg3YWpoODVFVU9zaG9KTWdSMGdaQ1FkRGFkaFh5S2xKQQpvZDcvV0JFWVhMa3ZSNWl1Tkk5aHl4Y3VYWkdEaWhNeUxLMmtXaWlIRDN0M3Zadjg4MjlJbDc5NURwTm9rNmhaCnUyS0F4UmtHR1U4aFBjR1hvVVF4eEdFQ0F3RUFBUT09Ci0tLS0tRU5EIFBVQkxJQyBLRVktLS0tLQ==',
                'fbc6d3d091d7c5ed745f499c548c103f03b96354c7bb3b3e982a128c',
                0,
                1,
                1,
                'genesis',
            )
        )
        hyp.commit()
        hyp.close()

        index = sqlite3.connect(index_path, timeout=1)
        index.text_factory = str
        index.execute('PRAGMA case_sensitive_like = 1;')
        index_cursor = index.cursor()
        index_cursor.execute('CREATE TABLE tokens (block_height INTEGER, timestamp, token, address, recipient, txid, amount INTEGER)')
        index_cursor.execute('CREATE TABLE aliases (block_height INTEGER, address, alias)')
        index_cursor.execute('CREATE INDEX "Alias Index" ON "aliases" ("block_height", "address","alias")')
        index_cursor.execute('CREATE INDEX "Token Index" ON "tokens" ("block_height","timestamp","token","address","recipient","txid","amount")')
        index_cursor.execute('CREATE TABLE staking (block_height INTEGER, timestamp NUMERIC, address, balance, ip, port, pos_address)')
        index.commit()
        index.close()

        time.sleep(2)
    except:
        traceback.print_exc()
        raise
    # print('Bootstrap successfully finished')


def check_db_for_bootstrap(node):
    upgrade = sqlite3.connect(node.ledger_path, timeout=1)
    u = upgrade.cursor()
    try:
        u.execute('SELECT * FROM transactions LIMIT 1;')
        result = u.fetchone()
        if not result:
            raise Exception()
        upgrade.close()
    except Exception as e:
        lg.warn(str(e))
        upgrade.close()
        lg.warn('Database needs upgrading, bootstrapping...')
        bootstrap()


#------------------------------------------------------------------------------


class CustomLogHandler(logging.Handler):

    def emit(self, record):
        try:
            if _Debug:
                lg.out(lg.get_debug_level() - 6, self.format(record))  # record.getMessage()
        except RecursionError:  # See issue 36272
            raise
        except Exception:
            self.handleError(record)

    def handleError(self, record):
        lg.err(self.format(record))


def custom_log(level_input='NOTSET'):
    if level_input == 'NOTSET':
        level = logging.NOTSET
    if level_input == 'DEBUG':
        level = logging.DEBUG
    if level_input == 'INFO':
        level = logging.INFO
    if level_input == 'WARNING':
        level = logging.WARNING
    if level_input == 'ERROR':
        level = logging.ERROR
    if level_input == 'CRITICAL':
        level = logging.CRITICAL

    log_formatter = logging.Formatter('%(module)s.%(funcName)s %(levelname)s %(message)s')
    my_handler = CustomLogHandler(level=level)
    my_handler.setFormatter(log_formatter)

    app_log = logging.getLogger('root')
    app_log.setLevel(level)
    app_log.addHandler(my_handler)

    return app_log



#!/usr/bin/env python
# p2p_service_seeker.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (p2p_service_seeker.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: p2p_service_seeker

.. role:: red

BitDust p2p_service_seeker() Automat

EVENTS:
    * :red:`connect-known-user`
    * :red:`fail`
    * :red:`found-users`
    * :red:`lookup-random-user`
    * :red:`request-timeout`
    * :red:`service-accepted`
    * :red:`service-denied`
    * :red:`shook-hands`
    * :red:`users-not-found`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

from twisted.internet.defer import Deferred

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.lib import strng
from bitdust.lib import packetid

from bitdust.main import settings

from bitdust.p2p import commands
from bitdust.p2p import p2p_service
from bitdust.p2p import handshaker

from bitdust.contacts import identitycache

from bitdust.userid import id_url
from bitdust.userid import global_id
from bitdust.userid import my_id

#------------------------------------------------------------------------------

_P2PServiceSeekerInstaceCounter = 0

#------------------------------------------------------------------------------


class P2PServiceSeeker(automat.Automat):

    """
    This class implements all the functionality of the ``p2p_service_seeker()``
    state machine.
    """

    fast = True

    def __repr__(self):
        return '%s[%s@%s%s%s](%s)' % (self.id, self.target_service or '', self.target_id or '', '#' if self.RandomLookup else '!', self.Attempts, self.state)

    def init(self):
        """
        Method to initialize additional variables and flags at creation phase
        of p2p_service_seeker() machine.
        """
        self.Attempts = 0
        self.RandomLookup = False
        self.lookup_method = None
        self.target_idurl = None
        self.target_id = None
        self.target_service = None
        self.request_service_params = None
        self.request_service_timeout = None
        self.ping_retries = None
        self.ack_timeout = None
        self.force_handshake = None
        self.exclude_nodes = []
        self.verify_accepted = None
        self.lookup_task = None
        self.requested_packet_id = None

    def A(self, event, *args, **kwargs):
        """
        The state machine code, generated using `visio2python
        <https://bitdust.io/visio2python/>`_ tool.
        """
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'connect-known-user':
                self.state = 'HANDSHAKE?'
                self.doInit(*args, **kwargs)
                self.Attempts = 0
                self.RandomLookup = False
                self.doSelectOneUser(*args, **kwargs)
                self.doHandshake(*args, **kwargs)
            elif event == 'lookup-random-user':
                self.state = 'RANDOM_USER?'
                self.doInit(*args, **kwargs)
                self.Attempts = 0
                self.RandomLookup = True
                self.doLookupRandomNode(*args, **kwargs)
        #---RANDOM_USER?---
        elif self.state == 'RANDOM_USER?':
            if event == 'found-users':
                self.state = 'HANDSHAKE?'
                self.doSelectOneUser(*args, **kwargs)
                self.Attempts += 1
                self.doHandshake(*args, **kwargs)
            elif event == 'users-not-found' and not self.isRetries(*args, **kwargs):
                self.state = 'FAILED'
                self.doNotifyLookupFailed(event, *args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'users-not-found' and self.isRetries(*args, **kwargs):
                self.Attempts += 1
                self.doLookupRandomNode(*args, **kwargs)
        #---HANDSHAKE?---
        elif self.state == 'HANDSHAKE?':
            if event == 'shook-hands':
                self.state = 'SERVICE?'
                self.doSendRequestService(*args, **kwargs)
            elif event == 'fail' and self.isRetries(*args, **kwargs) and self.RandomLookup:
                self.state = 'RANDOM_USER?'
                self.doLookupRandomNode(*args, **kwargs)
            elif (not self.isRetries(*args, **kwargs) or not self.RandomLookup) and event == 'fail':
                self.state = 'FAILED'
                self.doNotifyHandshakeFailed(event, *args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        #---SERVICE?---
        elif self.state == 'SERVICE?':
            if event == 'service-accepted':
                self.state = 'SUCCESS'
                self.doNotifyServiceAccepted(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif (event == 'request-timeout' or event == 'fail' or event == 'service-denied') and (not self.RandomLookup or (not self.isRetries(*args, **kwargs) and self.RandomLookup)):
                self.state = 'FAILED'
                self.doNotifyServiceRequestFailed(event, *args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif (event == 'request-timeout' or event == 'fail' or event == 'service-denied') and self.isRetries(*args, **kwargs) and self.RandomLookup:
                self.state = 'RANDOM_USER?'
                self.doLookupRandomNode(*args, **kwargs)
        #---SUCCESS---
        elif self.state == 'SUCCESS':
            pass
        #---FAILED---
        elif self.state == 'FAILED':
            pass
        return None

    def isRetries(self, *args, **kwargs):
        """
        Condition method.
        """
        return self.Attempts < self.retries

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        self.lookup_method = kwargs.get('lookup_method', None)
        self.target_service = kwargs['target_service']
        self.request_service_params = kwargs.get('request_service_params', None)
        self.request_service_timeout = kwargs.get('request_service_timeout', 120)
        self.ping_retries = kwargs.get('ping_retries', None)
        self.ack_timeout = kwargs.get('ack_timeout', None)
        self.force_handshake = kwargs.get('force_handshake', False)
        self.result_callback = kwargs.get('result_callback', None)
        self.verify_accepted = kwargs.get('verify_accepted', True)
        self.exclude_nodes = id_url.to_bin_list(kwargs.get('exclude_nodes', []))
        self.retries = kwargs.get('attempts', 5)

    def doLookupRandomNode(self, *args, **kwargs):
        """
        Action method.
        """
        self.lookup_task = self.lookup_method()
        if self.lookup_task.result_defer:
            self.lookup_task.result_defer.addCallback(self._nodes_lookup_finished)
            self.lookup_task.result_defer.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='p2p_service_seeker.doLookupRandomNode')
            self.lookup_task.result_defer.addErrback(lambda err: self.automat('users-not-found'))
        else:
            self.automat('users-not-found')

    def doSelectOneUser(self, *args, **kwargs):
        """
        Action method.
        """
        if 'remote_idurl' in kwargs:
            self.target_idurl = kwargs['remote_idurl']
        else:
            self.target_idurl = args[0][0]
        self.target_idurl = id_url.field(self.target_idurl)
        self.target_id = global_id.idurl2glob(self.target_idurl)

    def doHandshake(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.args(_DebugLevel, target_idurl=self.target_idurl)
        d = handshaker.ping(
            idurl=self.target_idurl,
            channel='p2p_service_seeker',
            keep_alive=True,
            force_cache=False,
            ping_retries=(1 if self.ping_retries is None else self.ping_retries),
            ack_timeout=(settings.P2PTimeOut() if self.ack_timeout is None else self.ack_timeout),
            cancel_running=self.force_handshake,
        )
        d.addCallback(lambda ok: self.automat('shook-hands'))
        d.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='p2p_service_seeker.doHandshake')
        d.addErrback(lambda err: self.automat('fail', reason='handshake-failed'))

    def doSendRequestService(self, *args, **kwargs):
        """
        Action method.
        """
        self.target_idurl.refresh()
        packet_id = packetid.UniqueID()
        if _Debug:
            lg.args(_DebugLevel, idurl=self.target_idurl, service=self.target_service, packet_id=packet_id)
        service_request_payload = self.request_service_params
        if callable(service_request_payload):
            service_request_payload = service_request_payload(self.target_idurl)
        out_packet = p2p_service.SendRequestService(
            remote_idurl=self.target_idurl,
            service_name=self.target_service,
            json_payload=service_request_payload,
            timeout=self.request_service_timeout,
            callbacks={
                commands.Ack(): self._node_acked,
                commands.Fail(): self._node_failed,
                None: self._node_timed_out,
            },
            packet_id=packet_id,
        )
        self.requested_packet_id = out_packet.PacketID

    def doNotifyServiceAccepted(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(_DebugLevel, 'p2p_service_seeker.doNotifyServiceAccepted %r from %r with %s' % (self.target_service, self.target_id, args[0]))
        if self.result_callback:
            self.result_callback('node-connected', *args, **kwargs)
        self.result_callback = None

    def doNotifyLookupFailed(self, event, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.args(_DebugLevel, event=event, attempts=self.Attempts, args=args, kwargs=kwargs)
        if self.result_callback:
            self.result_callback('lookup-failed', *args, **kwargs)
        self.result_callback = None

    def doNotifyServiceRequestFailed(self, event, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.args(_DebugLevel, event=event, attempts=self.Attempts, args=args, kwargs=kwargs)
        if self.result_callback:
            self.result_callback('request-failed', *args, **kwargs)
        self.result_callback = None

    def doNotifyHandshakeFailed(self, event, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.args(_DebugLevel, event=event, attempts=self.Attempts, args=args, kwargs=kwargs)
        if self.result_callback:
            self.result_callback('handshake-failed', *args, **kwargs)
        self.result_callback = None

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        self.lookup_method = None
        self.target_idurl = None
        self.target_id = None
        self.target_service = None
        self.request_service_params = None
        self.request_service_timeout = None
        self.ping_retries = None
        self.ack_timeout = None
        self.force_handshake = None
        self.exclude_nodes = []
        self.verify_accepted = None
        self.requested_packet_id = None
        self.lookup_task = None
        self.destroy()

    #------------------------------------------------------------------------------

    def _node_acked(self, response, info):
        if _Debug:
            lg.out(_DebugLevel, 'p2p_service_seeker._node_acked %r %r' % (response, info))
        if self.verify_accepted and not strng.to_text(response.Payload).startswith('accepted'):
            if _Debug:
                lg.out(_DebugLevel, 'p2p_service_seeker._node_acked with "service denied" response: %r %r' % (response, info))
            self.automat('service-denied', (response, info), reason='service-denied')
            return
        if _Debug:
            lg.out(_DebugLevel, 'p2p_service_seeker._node_acked %s is connected' % response.CreatorID)
        self.automat('service-accepted', (response, info))

    def _node_failed(self, response, info):
        if _Debug:
            lg.out(_DebugLevel, 'p2p_service_seeker._node_failed %r %r' % (response, info))
        self.automat('service-denied', (response, info), reason='service-denied')

    def _node_timed_out(self, pkt_out):
        if _Debug:
            lg.out(_DebugLevel, 'p2p_service_seeker._node_timed_out for outgoing packet %r' % pkt_out)
        self.automat('fail', pkt_out, reason='service-request-timeout')

    def _nodes_lookup_finished(self, idurls):
        if _Debug:
            lg.out(_DebugLevel, 'p2p_service_seeker._nodes_lookup_finished : %r' % idurls)
        found_idurls = []
        for idurl in idurls:
            if idurl in self.exclude_nodes:
                continue
            if idurl in found_idurls:
                continue
            ident = identitycache.FromCache(idurl)
            remoteprotos = set(ident.getProtoOrder())
            myprotos = set(my_id.getLocalIdentity().getProtoOrder())
            if len(myprotos.intersection(remoteprotos)) > 0:
                found_idurls.append(idurl)
        if found_idurls:
            self.automat('found-users', found_idurls)
        else:
            self.automat('users-not-found')


#------------------------------------------------------------------------------


def on_lookup_result(event, result_defer, *args, **kwargs):
    if _Debug:
        lg.args(_DebugLevel, event=event, args=args, kwargs=kwargs)
    if event == 'node-connected':
        result_defer.callback(*args, **kwargs)
    else:
        result_defer.errback(Exception(event, args, kwargs))


#------------------------------------------------------------------------------


def connect_random_node(lookup_method, service_name, service_params=None, verify_accepted=True, exclude_nodes=[], attempts=5, request_service_timeout=None, ping_retries=None, ack_timeout=None, force_handshake=False):
    global _P2PServiceSeekerInstaceCounter
    _P2PServiceSeekerInstaceCounter += 1
    result = Deferred()
    p2p_seeker = P2PServiceSeeker(
        name='p2p_service_seeker%d' % _P2PServiceSeekerInstaceCounter,
        state='AT_STARTUP',
        debug_level=_DebugLevel,
        log_events=_Debug,
        log_transitions=_Debug,
        publish_events=False,
    )
    p2p_seeker.automat(
        'lookup-random-user',
        lookup_method=lookup_method,
        target_service=service_name,
        request_service_params=service_params,
        request_service_timeout=request_service_timeout,
        ping_retries=ping_retries,
        attempts=attempts,
        ack_timeout=ack_timeout,
        force_handshake=force_handshake,
        verify_accepted=verify_accepted,
        result_callback=lambda evt, *a, **kw: on_lookup_result(evt, result, *a, **kw),
        exclude_nodes=exclude_nodes,
    )
    if _Debug:
        lg.args(_DebugLevel, service_name=service_name, exclude_nodes=exclude_nodes, inst=p2p_seeker)
    return result


def connect_known_node(remote_idurl, service_name, service_params=None, verify_accepted=True, exclude_nodes=[], attempts=2, request_service_timeout=None, ping_retries=None, ack_timeout=None, force_handshake=False):
    global _P2PServiceSeekerInstaceCounter
    _P2PServiceSeekerInstaceCounter += 1
    result = Deferred()
    p2p_seeker = P2PServiceSeeker(
        name='p2p_service_seeker%d' % _P2PServiceSeekerInstaceCounter,
        state='AT_STARTUP',
        debug_level=_DebugLevel,
        log_events=_Debug,
        log_transitions=_Debug,
        publish_events=False,
    )
    p2p_seeker.automat(
        'connect-known-user',
        remote_idurl=remote_idurl,
        target_service=service_name,
        request_service_params=service_params,
        request_service_timeout=request_service_timeout,
        ping_retries=ping_retries,
        attempts=attempts,
        ack_timeout=ack_timeout,
        force_handshake=force_handshake,
        verify_accepted=verify_accepted,
        result_callback=lambda evt, *a, **kw: on_lookup_result(evt, result, *a, **kw),
        exclude_nodes=exclude_nodes,
    )
    if _Debug:
        lg.args(_DebugLevel, service_name=service_name, exclude_nodes=exclude_nodes, inst=p2p_seeker)
    return result



#!/usr/bin/python
# service_identity_propagate.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_identity_propagate.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_identity_propagate
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return IdentityPropagateService()


class IdentityPropagateService(LocalService):

    service_name = 'service_identity_propagate'
    config_path = 'services/identity-propagate/enabled'

    def dependent_on(self):
        return [
            'service_gateway',
            'service_tcp_connections',
        ]

    def installed(self):
        from bitdust.userid import my_id
        if not my_id.isLocalIdentityReady():
            return False
        return True

    def network_configuration(self):
        import re
        from bitdust.main import config
        known_identity_servers_str = str(config.conf().getString('services/identity-propagate/known-servers'))
        known_identity_servers = []
        for id_server_str in re.split('\n|;|,| ', known_identity_servers_str):
            if id_server_str.strip():
                try:
                    id_server = id_server_str.strip().split(':')
                    id_server_host = id_server[0].strip()
                    id_server_http_port = int(id_server[1].strip())
                    if len(id_server) > 2:
                        id_server_tcp_port = int(id_server[2].strip())
                    else:
                        id_server_tcp_port = 6661
                except:
                    continue
                known_identity_servers.append({
                    'host': id_server_host,
                    'tcp_port': id_server_tcp_port,
                    'http_port': id_server_http_port,
                })
        if not known_identity_servers:
            from bitdust.main import network_config
            default_network_config = network_config.read_network_config_file()
            known_identity_servers = default_network_config['service_identity_propagate']['known_servers']
        return {
            'known_servers': known_identity_servers,
            'whitelisted_servers': [],
        }

    def start(self):
        from bitdust.logs import lg
        from bitdust.userid import my_id
        from bitdust.main.config import conf
        my_id.loadLocalIdentity()
        if not my_id.isLocalIdentityReady():
            lg.warn('Loading local identity failed - need to create an identity first')
            return False
        from bitdust.main import events
        from bitdust.contacts import identitycache
        from bitdust.userid import known_servers
        from bitdust.p2p import propagate
        from bitdust.contacts import contactsdb
        identitycache.init()
        d = contactsdb.init()
        propagate.init()
        conf().addConfigNotifier('services/identity-propagate/known-servers', self._on_known_servers_changed)
        lg.info('known ID servers are : %r' % known_servers.by_host())
        events.add_subscriber(self._on_local_identity_modified, 'local-identity-modified')
        events.add_subscriber(self._on_my_identity_rotated, 'my-identity-rotated')
        return d

    def stop(self):
        from bitdust.main.config import conf
        from bitdust.main import events
        from bitdust.p2p import propagate
        from bitdust.contacts import contactsdb
        from bitdust.contacts import identitycache
        events.remove_subscriber(self._on_my_identity_rotated, 'my-identity-rotated')
        events.remove_subscriber(self._on_local_identity_modified, 'local-identity-modified')
        conf().removeConfigNotifier('services/identity-propagate/known-servers')
        propagate.shutdown()
        contactsdb.shutdown()
        identitycache.shutdown()
        return True

    def _on_known_servers_changed(self, path, value, oldvalue, result):
        from bitdust.userid import known_servers
        known_servers._KnownServers = None

    def _on_local_identity_modified(self, evt):
        from bitdust.p2p import propagate
        propagate.update()

    def _on_my_identity_rotated(self, evt):
        from bitdust.logs import lg
        from bitdust.p2p import propagate
        from bitdust.contacts import contactsdb
        known_remote_contacts = set(contactsdb.contacts_remote(include_all=True))
        propagate.startup_list().update(known_remote_contacts)
        lg.warn('added %d known contacts to propagate startup list to be sent later' % len(known_remote_contacts))



#!/usr/bin/python
# backup_monitor.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (backup_monitor.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: backup_monitor.

.. raw:: html

    <a href="https://bitdust.io/automats/backup_monitor/backup_monitor.png" target="_blank">
    <img src="https://bitdust.io/automats/backup_monitor/backup_monitor.png" style="max-width:100%;">
    </a>

This is a state machine to manage rebuilding process of all backups.

Do several operations periodically:
    1) ping all suppliers
    2) request ListFiles from all suppliers
    3) prepare a list of backups to put some work to rebuild
    4) run rebuilding process and wait to finish
    5) make decision to replace one unreliable supplier with fresh one

The ``backup_monitor()`` automat starts the process of rebuilding the backups.

Control is passed to the ``list_files_orator()`` machine,
which will update the list of user's files already stored on remote machines.

Next would be a perform a list of backups that need to be rebuilt.

In the next step, control is passed to the state machine ``backup_rebuilder()``,
which control of the rebuilding process.

The last step is run ``fire_hire()`` automat, which monitors remote suppliers.


EVENTS:
    * :red:`backup_rebuilder.state`
    * :red:`fire-hire-finished`
    * :red:`init`
    * :red:`instant`
    * :red:`list-backups-done`
    * :red:`list_files_orator.state`
    * :red:`restart`
    * :red:`suppliers-changed`
    * :red:`timer-5sec`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 12

#------------------------------------------------------------------------------

import gc
import sys
import time

#------------------------------------------------------------------------------

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor in backup_monitor.py')

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.lib import diskspace
from bitdust.lib import packetid
from bitdust.lib import misc

from bitdust.main import settings

from bitdust.contacts import contactsdb

from bitdust.storage import backup_matrix
from bitdust.storage import backup_fs
from bitdust.storage import backup_control

from bitdust.userid import global_id
from bitdust.userid import my_id

from bitdust.p2p import online_status

#------------------------------------------------------------------------------

_BackupMonitor = None

#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with the state machine.
    """
    global _BackupMonitor
    if _BackupMonitor is None:
        _BackupMonitor = BackupMonitor(
            'backup_monitor',
            'AT_STARTUP',
            debug_level=_DebugLevel,
            log_events=False,
            log_transitions=_Debug,
        )
    if event is not None:
        _BackupMonitor.automat(event, *args, **kwargs)
    return _BackupMonitor


def Destroy():
    """
    Destroy backup_monitor() automat and remove its instance from memory.
    """
    global _BackupMonitor
    if _BackupMonitor is None:
        return
    _BackupMonitor.destroy()
    del _BackupMonitor
    _BackupMonitor = None


class BackupMonitor(automat.Automat):

    """
    A class to monitor backups and manage rebuilding process.
    """

    timers = {
        'timer-5sec': (5.0, ['READY']),
    }

    def init(self):
        self.current_suppliers = []
        self.backups_progress_last_iteration = 0
        self.last_execution_time = 0

    def state_changed(self, oldstate, newstate, event, *args, **kwargs):
        """
        This method is called every time when my state is changed.
        """
        if newstate == 'READY' and event != 'instant':
            self.automat('instant')

    def A(self, event, *args, **kwargs):
        from bitdust.customer import fire_hire
        from bitdust.customer import list_files_orator
        from bitdust.storage import backup_rebuilder
        from bitdust.storage import index_synchronizer
        from bitdust.stream import data_sender
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'READY'
                self.RestartAgain = False
        #---READY---
        elif self.state == 'READY':
            if event == 'timer-5sec':
                self.doOverallCheckUp(*args, **kwargs)
            elif event == 'restart' or event == 'suppliers-changed' or (event == 'instant' and self.RestartAgain):
                self.state = 'FIRE_HIRE'
                self.RestartAgain = False
                self.doRememberSuppliers(*args, **kwargs)
                fire_hire.A('restart')
        #---LIST_FILES---
        elif self.state == 'LIST_FILES':
            if (event == 'list_files_orator.state' and args[0] == 'NO_FILES'):
                self.state = 'READY'
            elif (event == 'list_files_orator.state' and args[0] == 'SAW_FILES'):
                self.state = 'LIST_BACKUPS'
                index_synchronizer.A('pull')
                data_sender.A('restart')
                self.doPrepareListBackups(*args, **kwargs)
            elif event == 'restart':
                self.RestartAgain = True
            elif event == 'suppliers-changed':
                self.state = 'READY'
                self.RestartAgain = True
        #---LIST_BACKUPS---
        elif self.state == 'LIST_BACKUPS':
            if event == 'list-backups-done':
                self.state = 'REBUILDING'
                backup_rebuilder.A('start')
            elif event == 'restart':
                self.RestartAgain = True
            elif event == 'suppliers-changed':
                self.state = 'READY'
                self.RestartAgain = True
            elif event == 'restart':
                self.state = 'FIRE_HIRE'
                fire_hire.A('restart')
        #---REBUILDING---
        elif self.state == 'REBUILDING':
            if (event == 'backup_rebuilder.state' and args[0] in ['DONE', 'STOPPED']):
                self.state = 'READY'
                self.doCleanUpBackups(*args, **kwargs)
                data_sender.A('restart')
            elif event == 'restart' or event == 'suppliers-changed':
                self.state = 'FIRE_HIRE'
                backup_rebuilder.SetStoppedFlag()
                fire_hire.A('restart')
        #---FIRE_HIRE---
        elif self.state == 'FIRE_HIRE':
            if event == 'suppliers-changed' and self.isSuppliersNumberChanged(*args, **kwargs):
                self.state = 'LIST_FILES'
                self.doDeleteAllBackups(*args, **kwargs)
                self.doRememberSuppliers(*args, **kwargs)
                list_files_orator.A('need-files')
            elif event == 'fire-hire-finished':
                self.state = 'LIST_FILES'
                list_files_orator.A('need-files')
            elif event == 'suppliers-changed' and not self.isSuppliersNumberChanged(*args, **kwargs):
                self.state = 'LIST_FILES'
                self.doUpdateSuppliers(*args, **kwargs)
                self.doRememberSuppliers(*args, **kwargs)
                list_files_orator.A('need-files')
            elif event == 'restart':
                self.RestartAgain = True
        return None

    def isSuppliersNumberChanged(self, *args, **kwargs):
        """
        Condition method.
        """
        return contactsdb.num_suppliers() != len(self.current_suppliers)

    def doRememberSuppliers(self, *args, **kwargs):
        """
        Action method.
        """
        self.current_suppliers = list(contactsdb.suppliers())
        self.backups_progress_last_iteration = 0
        self.last_execution_time = time.time()

    def doDeleteAllBackups(self, *args, **kwargs):
        """
        Action method.
        """
        lg.warn('about to erase all my backups')
        # cancel all tasks and jobs
        backup_control.DeleteAllTasks()
        backup_control.AbortAllRunningBackups()
        # remove all local files and all backups
        backup_control.DeleteAllBackups()
        # erase all remote info
        backup_matrix.ClearRemoteInfo()
        # also erase local info
        backup_matrix.ClearLocalInfo()
        # finally save the list of current suppliers and clear all stats
        from bitdust.stream import io_throttle
        io_throttle.DeleteAllSuppliers()

    def doUpdateSuppliers(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.stream import io_throttle
        # take a list of suppliers positions that ware changed
        changedSupplierNums = backup_matrix.SuppliersChangedNumbers(self.current_suppliers)
        # notify io_throttle that we do not need already those suppliers
        for supplierNum in changedSupplierNums:
            suplier_idurl = self.current_suppliers[supplierNum]
            if suplier_idurl:
                io_throttle.DeleteSuppliers([
                    suplier_idurl,
                ])
            # erase (set to 0) remote info for this guys
            backup_matrix.ClearSupplierRemoteInfo(supplierNum)

    def doPrepareListBackups(self, *args, **kwargs):
        from bitdust.storage import backup_rebuilder
        if backup_control.HasRunningBackup():
            # if some backups are running right now no need to rebuild something - too much use of CPU
            backup_rebuilder.RemoveAllBackupsToWork()
            if _Debug:
                lg.out(_DebugLevel, 'backup_monitor.doPrepareListBackups skip all rebuilds')
            self.automat('list-backups-done')
            return
        # take remote and local backups and get union from it
        allBackupIDs = set(list(backup_matrix.local_files().keys()) + list(backup_matrix.remote_files().keys()))
        # take only backups from data base
        allBackupIDs.intersection_update(backup_fs.ListAllBackupIDs(customer_idurl=my_id.getIDURL()))
        # remove running backups
        allBackupIDs.difference_update(backup_control.ListRunningBackups())
        # sort it in reverse order - newer backups should be repaired first
        allBackupIDs = misc.sorted_backup_ids(list(allBackupIDs), True)
        # add backups to the queue
        backup_rebuilder.AddBackupsToWork(allBackupIDs)
        if _Debug:
            lg.out(_DebugLevel, 'backup_monitor.doPrepareListBackups %d items:' % len(allBackupIDs))
        self.automat('list-backups-done', allBackupIDs)

    def doCleanUpBackups(self, *args, **kwargs):
        # here we check all backups we have and remove the old one
        # user can set how many versions of that file or folder to keep
        # other versions (older) will be removed here
        from bitdust.storage import backup_rebuilder
        try:
            self.backups_progress_last_iteration = len(backup_rebuilder.A().backupsWasRebuilt)
        except:
            self.backups_progress_last_iteration = 0
        versionsToKeep = settings.getBackupsMaxCopies()
        if not contactsdb.num_suppliers():
            bytesUsed = 0
        else:
            bytesUsed = backup_fs.sizebackups()/contactsdb.num_suppliers()
        bytesNeeded = diskspace.GetBytesFromString(settings.getNeededString(), 0)
        if _Debug:
            lg.out(_DebugLevel, 'backup_monitor.doCleanUpBackups backupsToKeep=%d used=%d needed=%d' % (versionsToKeep, bytesUsed, bytesNeeded))
        delete_count = 0
        if versionsToKeep > 0:
            for pathID, localPath, itemInfo in backup_fs.IterateIDs():
                pathID = global_id.CanonicalID(pathID)
                if backup_control.IsPathInProcess(pathID):
                    continue
                versions = itemInfo.list_versions(True, False)
                if _Debug:
                    lg.out(_DebugLevel, 'backup_monitor.doCleanUpBackups %r %r' % (pathID, len(versions)))
                while len(versions) > versionsToKeep:
                    oldest_version = versions.pop(0)
                    backupID = packetid.MakeBackupID(path_id=pathID, version=oldest_version, normalize_key_alias=False)
                    if _Debug:
                        lg.out(_DebugLevel, 'backup_monitor.doCleanUpBackups %d of %d backups for %s, so remove older %s' % (len(versions) + 1, versionsToKeep, pathID, oldest_version))
                    backup_control.DeleteBackup(backupID, saveDB=False, calculate=False)
                    delete_count += 1
        # we need also to fit used space into needed space (given from other users)
        # they trust us - do not need to take extra space from our friends
        # so remove oldest backups, but keep at least one for every folder - at least locally!
        # still our suppliers will remove our "extra" files by their "local_tester"
        if bytesNeeded <= bytesUsed:
            sizeOk = False
            for pathID, localPath, itemInfo in backup_fs.IterateIDs():
                if sizeOk:
                    break
                pathID = global_id.CanonicalID(pathID)
                versions = itemInfo.list_versions(True, False)
                if len(versions) <= 1:
                    continue
                for version in versions[1:]:
                    backupID = packetid.MakeBackupID(path_id=pathID, version=version, normalize_key_alias=False)
                    versionInfo = itemInfo.get_version_info(version)
                    if versionInfo[1] > 0:
                        if _Debug:
                            lg.out(_DebugLevel, 'backup_monitor.doCleanUpBackups over use %d of %d, so remove %s of %s' % (bytesUsed, bytesNeeded, backupID, localPath))
                        backup_control.DeleteBackup(backupID, saveDB=False, calculate=False)
                        delete_count += 1
                        bytesUsed -= versionInfo[1]
                        if bytesNeeded > bytesUsed:
                            sizeOk = True
                            break
        if delete_count > 0:
            backup_fs.Scan()
            backup_fs.Calculate()
            backup_control.SaveFSIndex()
        collected = gc.collect()
        if self.backups_progress_last_iteration > 0:
            if _Debug:
                lg.out(_DebugLevel, 'backup_monitor.doCleanUpBackups  sending "restart", backups_progress_last_iteration=%s' % self.backups_progress_last_iteration)
            reactor.callLater(1, self.automat, 'restart')  # @UndefinedVariable
        if _Debug:
            lg.out(_DebugLevel, 'backup_monitor.doCleanUpBackups delete_count=%r GC collected=%d' % (delete_count, collected))

    def doOverallCheckUp(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.customer import fire_hire
        if not fire_hire.IsAllHired():
            if _Debug:
                lg.out(_DebugLevel, 'backup_monitor.doOverallCheckUp some suppliers not hired yet, restart now')
            self.automat('restart')
            return
        if online_status.listOfflineSuppliers():
            if time.time() - self.last_execution_time > 60:
                # re-sync every 1 min. if at least on supplier is dead
                if _Debug:
                    lg.out(_DebugLevel, 'backup_monitor.doOverallCheckUp   restart after 1 min, found offline suppliers')
                self.automat('restart')
                return
        if time.time() - self.last_execution_time > 60*10:
            # also re-sync every 10 min.
            if _Debug:
                lg.out(_DebugLevel, 'backup_monitor.doOverallCheckUp   periodic 10 min. restart')
            self.automat('restart')
            return
        # TODO: more tests here: low rating(), time offline, low ping, etc..



#!/usr/bin/env python
# jsontemplate.py
#
# Copyright (C) 2009 Andy Chu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Python implementation of json-template.

JSON Template is a minimal and powerful templating language for transforming a
JSON dictionary to arbitrary text.

To use this module, you will typically use the Template constructor, and catch
various exceptions thrown.  You may also want to use the FromFile/FromString
methods, which allow Template constructor options to be embedded in the template
string itself.

Other functions are exposed for tools which may want to process templates.
"""

from __future__ import absolute_import
import six
from six.moves import range

__author__ = 'Andy Chu'

__all__ = [
    'Error', 'CompilationError', 'EvaluationError', 'BadFormatter', 'BadPredicate', 'MissingFormatter', 'ConfigurationError', 'TemplateSyntaxError', 'UndefinedVariable', 'CompileTemplate', 'FromString', 'FromFile', 'Template', 'expand'
]

from io import StringIO
import pprint
import re

# For formatters
import six.moves.urllib.parse  # for urllib.encode
import six.moves.urllib.parse  # for urljoin

if not six.PY2:
    from html import escape as html_escape
else:
    from cgi import escape as html_escape


class Error(Exception):
    """
    Base class for all exceptions in this module.

    Thus you can "except jsontemplate.Error: to catch all exceptions
    thrown by this module.
    """
    def __str__(self):
        """
        This helps people debug their templates.

        If a variable isn't defined, then some context is shown in the
        traceback. TODO: Attach context for other errors.
        """
        if hasattr(self, 'near'):
            return '%s\n\nNear: %s' % (self.args[0], pprint.pformat(self.near))
        else:
            return self.args[0]


class CompilationError(Error):
    """
    Base class for errors that happen during the compilation stage.
    """


class EvaluationError(Error):
    """
    Base class for errors that happen when expanding the template.

    This class of errors generally involve the data dictionary or the
    execution of the formatters.
    """
    def __init__(self, msg, original_exception=None):
        Error.__init__(self, msg)
        self.original_exception = original_exception


class BadFormatter(CompilationError):
    """
    A bad formatter was specified, e.g. {variable|BAD}
    """


class BadPredicate(CompilationError):
    """
    A bad predicate was specified, e.g. {.BAD?}
    """


class MissingFormatter(CompilationError):
    """
    Raised when formatters are required, and a variable is missing a formatter.
    """


class ConfigurationError(CompilationError):
    """
    Raised when the Template options are invalid and it can't even be compiled.
    """


class TemplateSyntaxError(CompilationError):
    """
    Syntax error in the template text.
    """


class UndefinedVariable(EvaluationError):
    """
    The template contains a variable not defined by the data dictionary.
    """


_SECTION_RE = re.compile(r'(repeated)?\s*section\s+(\S+)')

# Some formatters and predicates need to look up values in the whole context,
# rather than just the current node.  'Node functions' start with a lowercase
# letter; 'Context functions' start with any other character.
SIMPLE_FUNC, ENHANCED_FUNC = 0, 1


class FunctionRegistry(object):
    """
    Abstract class for looking up formatters or predicates at compile time.
    """
    def Lookup(self, user_str):
        """
        Lookup a function.

        Args:
          user_str: A raw string from the user, which may include uninterpreted
            arguments.  For example, 'pluralize person people' or 'test? admin'

        Returns:
          A 2-tuple of (function, args)
            function: Callable that formats data as a string
            args: Extra arguments to be passed to the function at expansion time
              Should be None to pass NO arguments, since it can pass a 0-tuple too.
        """
        raise NotImplementedError

    def LookupWithType(self, user_str):
        func, args = self.Lookup(user_str)
        # If users need the complexity of FunctionRegistry, then they get the
        # 3-arguments formatter signature (value, context, args)
        return func, args, ENHANCED_FUNC


def _DecideFuncType(user_str):
    """
    By default, formatters/predicates which start with a non-lowercase letter
    take contexts rather than just the cursor.
    """
    if user_str[0].islower():
        return SIMPLE_FUNC
    else:
        return ENHANCED_FUNC


class DictRegistry(FunctionRegistry):
    """
    Look up functions in a simple dictionary.
    """
    def __init__(self, func_dict):
        self.__dict__ = func_dict

    def LookupWithType(self, user_str):
        return self.__dict__.get(user_str), None, _DecideFuncType(user_str)


class CallableRegistry(FunctionRegistry):
    """
    Look up functions in a (higher-order) function.
    """
    def __init__(self, func):
        self.func = func

    def LookupWithType(self, user_str):
        return self.func(user_str), None, _DecideFuncType(user_str)


class PrefixRegistry(FunctionRegistry):
    """
    Lookup functions with arguments.

    The function name is identified by a prefix.  The character after
    the prefix, usually a space, is considered the argument delimiter
    (similar to sed/perl's s/foo/bar s|foo|bar syntax).
    """
    def __init__(self, functions):
        """
        Args: functions: List of 2-tuples (prefix, function), e.g.

        [('pluralize', _Pluralize), ('cycle', _Cycle)]
        """
        self.functions = functions

    def Lookup(self, user_str):
        for prefix, func in self.functions:
            if user_str.startswith(prefix):
                i = len(prefix)

                # Usually a space, but could be something else
                try:
                    splitchar = user_str[i]
                except IndexError:
                    args = ()  # No arguments
                else:
                    args = user_str.split(splitchar)[1:]

                return func, args
        return None, ()


class ChainedRegistry(FunctionRegistry):
    """
    Look up functions in chain of other FunctionRegistry instances.
    """
    def __init__(self, registries):
        self.registries = registries

    def LookupWithType(self, user_str):
        for registry in self.registries:
            func, args, func_type = registry.LookupWithType(user_str)
            if func:
                return func, args, func_type

        # Nothing found
        return None, None, SIMPLE_FUNC


class _ProgramBuilder(object):
    """
    Receives method calls from the parser, and constructs a tree of _Section()
    instances.
    """
    def __init__(self, formatters, predicates):
        """
        Args:

        formatters: See docstring for CompileTemplate predicates: See
        docstring for CompileTemplate
        """
        self.current_block = _Section()
        self.stack = [self.current_block]

        # Passing a dictionary or a function is often more convenient than making a
        # FunctionRegistry
        if isinstance(formatters, dict):
            formatters = DictRegistry(formatters)
        elif callable(formatters):
            formatters = CallableRegistry(formatters)

        # default formatters with arguments
        default_formatters = PrefixRegistry([('pluralize', _Pluralize), ('cycle', _Cycle)])

        # First consult user formatters, then the default formatters
        self.formatters = ChainedRegistry([formatters, DictRegistry(_DEFAULT_FORMATTERS), default_formatters])

        # Same for predicates
        if isinstance(predicates, dict):
            predicates = DictRegistry(predicates)
        elif callable(predicates):
            predicates = CallableRegistry(predicates)

        self.predicates = ChainedRegistry([predicates, DictRegistry(_DEFAULT_PREDICATES)])

    def Append(self, statement):
        """
        Args:

        statement: Append a literal
        """
        self.current_block.Append(statement)

    def _GetFormatter(self, format_str):
        """
        The user's formatters are consulted first, then the default formatters.
        """
        formatter, args, func_type = self.formatters.LookupWithType(format_str)
        if formatter:
            return formatter, args, func_type
        else:
            raise BadFormatter('%r is not a valid formatter' % format_str)

    def _GetPredicate(self, pred_str):
        """
        The user's predicates are consulted first, then the default predicates.
        """
        predicate, args, func_type = self.predicates.LookupWithType(pred_str)
        if predicate:
            return predicate, args, func_type
        else:
            raise BadPredicate('%r is not a valid predicate' % pred_str)

    def AppendSubstitution(self, name, formatters):
        formatters = [self._GetFormatter(f) for f in formatters]
        self.current_block.Append((_DoSubstitute, (name, formatters)))

    def _NewSection(self, func, new_block):
        self.current_block.Append((func, new_block))
        self.stack.append(new_block)
        self.current_block = new_block

    def NewSection(self, token_type, section_name):
        """
        For sections or repeated sections.
        """

        # TODO: Consider getting rid of this dispatching, and turn _Do* into methods
        if token_type == REPEATED_SECTION_TOKEN:
            new_block = _RepeatedSection(section_name)
            func = _DoRepeatedSection
        elif token_type == SECTION_TOKEN:
            new_block = _Section(section_name)
            func = _DoSection
        else:
            raise AssertionError('Invalid token type %s' % token_type)

        self._NewSection(func, new_block)

    def NewOrClause(self, pred_str):
        """
        {.or ...} Can appear inside predicate blocks or section blocks, with
        slightly different meaning.
        """
        if pred_str:
            pred = self._GetPredicate(pred_str)
        else:
            pred = None
        self.current_block.NewOrClause(pred)

    def AlternatesWith(self):
        self.current_block.AlternatesWith()

    def NewPredicateSection(self, pred_str):
        """
        For chains of predicate clauses.
        """
        pred = self._GetPredicate(pred_str)
        block = _PredicateSection()
        block.NewOrClause(pred)

        self._NewSection(_DoPredicates, block)

    def EndSection(self):
        self.stack.pop()
        self.current_block = self.stack[-1]

    def Root(self):
        # It's assumed that we call this at the end of the program
        return self.current_block


class _AbstractSection(object):
    def __init__(self):
        # Pairs of func, args, or a literal string
        self.current_clause = []

    def Append(self, statement):
        """
        Append a statement to this block.
        """
        self.current_clause.append(statement)

    def AlternatesWith(self):
        raise TemplateSyntaxError('{.alternates with} can only appear with in {.repeated section ...}')

    def NewOrClause(self):
        raise NotImplementedError


class _Section(_AbstractSection):
    """
    Represents a (repeated) section.
    """
    def __init__(self, section_name=None):
        """
        Args:

        section_name: name given as an argument to the section
        token_type: The token type that created this section (e.g.
        PREDICATE_TOKEN)
        """
        _AbstractSection.__init__(self)
        self.section_name = section_name

        # Clauses is just a string and a list of statements.
        self.statements = {'default': self.current_clause}

    def __repr__(self):
        return '<Section %s>' % self.section_name

    def Statements(self, clause='default'):
        return self.statements.get(clause, [])

    def NewOrClause(self, pred):
        if pred:
            raise TemplateSyntaxError('{.or} clause only takes a predicate inside predicate blocks')
        self.current_clause = []
        self.statements['or'] = self.current_clause


class _RepeatedSection(_Section):
    """
    Repeated section is like section, but it supports {.alternates with}
    """
    def AlternatesWith(self):
        self.current_clause = []
        self.statements['alternates with'] = self.current_clause


class _PredicateSection(_AbstractSection):
    """
    Represents a sequence of predicate clauses.
    """
    def __init__(self):
        _AbstractSection.__init__(self)
        # List of func, statements
        self.clauses = []

    def NewOrClause(self, pred):
        # {.or} always executes if reached
        pred = pred or (lambda x: True, None, SIMPLE_FUNC)  # 3-tuple
        self.current_clause = []
        self.clauses.append((pred, self.current_clause))


class _Frame(object):
    """
    A stack frame.
    """
    def __init__(self, context, index=-1):
        # Public attributes
        self.context = context
        self.index = index  # An iteration index.  -1 means we're NOT iterating.

    def __str__(self):
        return 'Frame %s (%s)' % (self.context, self.index)


class _ScopedContext(object):
    """
    Allows scoped lookup of variables.

    If the variable isn't in the current context, then we search up the
    stack.
    """
    def __init__(self, context, undefined_str):
        """
        Args:

        context: The root context undefined_str: See Template()
        constructor.
        """
        self.stack = [_Frame(context)]
        self.undefined_str = undefined_str

    def PushSection(self, name):
        """
        Given a section name, push it on the top of the stack.

        Returns:   The new section, or None if there is no such section.
        """
        if name == '@':
            new_context = self.stack[-1].context
        else:
            new_context = self.stack[-1].context.get(name)
        self.stack.append(_Frame(new_context))
        return new_context

    def Pop(self):
        self.stack.pop()

    def Next(self):
        """
        Advance to the next item in a repeated section.

        Raises:   StopIteration if there are no more elements
        """
        stacktop = self.stack[-1]

        # Now we're iterating -- push a new mutable object onto the stack
        if stacktop.index == -1:
            stacktop = _Frame(None, index=0)
            self.stack.append(stacktop)

        context_array = self.stack[-2].context

        if stacktop.index == len(context_array):
            self.stack.pop()
            raise StopIteration

        stacktop.context = context_array[stacktop.index]
        stacktop.index += 1

        return True  # OK, we mutated the stack

    def _Undefined(self, name):
        if self.undefined_str is None:
            raise UndefinedVariable('%r is not defined' % name)
        else:
            return self.undefined_str

    def _LookUpStack(self, name):
        """
        Look up the stack for the given name.
        """
        i = len(self.stack) - 1
        while True:
            frame = self.stack[i]
            if name == '@index':
                if frame.index != -1:  # -1 is undefined
                    return frame.index  # @index is 1-based
            else:
                context = frame.context
                if hasattr(context, 'get'):  # Can't look up names in a list or atom
                    try:
                        return context[name]
                    except KeyError:
                        pass

            i -= 1  # Next frame
            if i <= -1:  # Couldn't find it anywhere
                return self._Undefined(name)

    def Lookup(self, name):
        """
        Get the value associated with a name in the current context.

        The current context could be an dictionary in a list, or a dictionary
        outside a list.

        Args:
          name: name to lookup, e.g. 'foo' or 'foo.bar.baz'

        Returns:
          The value, or self.undefined_str

        Raises:
          UndefinedVariable if self.undefined_str is not set
        """
        if name == '@':
            return self.stack[-1].context

        parts = name.split('.')
        value = self._LookUpStack(parts[0])

        # Now do simple lookups of the rest of the parts
        for part in parts[1:]:
            try:
                value = value[part]
            except (KeyError, TypeError):  # TypeError for non-dictionaries
                return self._Undefined(part)

        return value


def _ToString(x):
    # Some cross-language values for primitives
    if x is None:
        return 'null'
    if isinstance(x, six.string_types):
        return x
    return pprint.pformat(x)


def _HtmlAttrValue(x):
    return html_escape(x, quote=True)


def _AbsUrl(relative_url, context, unused_args):
    """
    Returns an absolute URL, given the current node as a relative URL.

    Assumes that the context has a value named 'base-url'.  This is a little like
    the HTML <base> tag, but implemented with HTML generation.

    Raises:
      UndefinedVariable if 'base-url' doesn't exist
    """
    # urljoin is flexible about trailing/leading slashes -- it will add or de-dupe
    # them
    return six.moves.urllib.parse.urljoin(context.Lookup('base-url'), relative_url)


# See http://google-ctemplate.googlecode.com/svn/trunk/doc/howto.html for more
# escape types.
#
# Also, we might want to take a look at Django filters.
#
# This is a *public* constant, so that callers can use it construct their own
# formatter lookup dictionaries, and pass them in to Template.
_DEFAULT_FORMATTERS = {
    'html': html_escape,

    # The 'htmltag' name is deprecated.  The html-attr-value name is preferred
    # because it can be read with "as":
    #   {url|html-attr-value} means:
    #   "substitute 'url' as an HTML attribute value"
    'html-attr-value': _HtmlAttrValue,
    'htmltag': _HtmlAttrValue,
    'raw': lambda x: x,  # Used for the length of a list.  Can be used for the size of a dictionary
    # too, though I haven't run into that use case.
    'size': lambda value: str(len(value)),

    # The argument is a dictionary, and we get a a=1&b=2 string back.
    'url-params': six.moves.urllib.parse.urlencode,

    # The argument is an atom, and it takes 'Search query?' -> 'Search+query%3F'
    'url-param-value': six.moves.urllib.parse.quote_plus,  # param is an atom

    # The default formatter, when no other default is specifier.  For debugging,
    # this could be lambda x: json.dumps(x, indent=2), but here we want to be
    # compatible to Python 2.4.
    'str': _ToString,

    # Just show a plain URL on an HTML page (without anchor text).
    'plain-url': lambda x: '<a href="%s">%s</a>' % (html_escape(x, quote=True), html_escape(x)),

    # A context formatter
    'AbsUrl': _AbsUrl,

    # Placeholders for "standard names".  We're not including them by default
    # since they require additional dependencies.  We can provide a part of the
    # "lookup chain" in formatters.py for people people want the dependency.

    # 'json' formats arbitrary data dictionary nodes as JSON strings.  'json'
    # and 'js-string' are identical (since a JavaScript string *is* JSON).  The
    # latter is meant to be serve as extra documentation when you want a string
    # argument only, which is a common case.
    'json': None,
    'js-string': None,
}


def _Pluralize(value, unused_context, args):
    """
    Formatter to pluralize words.
    """

    if len(args) == 0:
        s, p = '', 's'
    elif len(args) == 1:
        s, p = '', args[0]
    elif len(args) == 2:
        s, p = args
    else:
        # Should have been checked at compile time
        raise AssertionError

    if value > 1:
        return p
    else:
        return s


def _Cycle(value, unused_context, args):
    """
    Cycle between various values on consecutive integers.
    """
    # @index starts from 1, so used 1-based indexing
    return args[(value - 1) % len(args)]


def _IsDebugMode(unused_value, context, unused_args):
    try:
        return bool(context.Lookup('debug'))
    except UndefinedVariable:
        return False


_DEFAULT_PREDICATES = {
    'singular?': lambda x: x == 1,
    'plural?': lambda x: x > 1,
    'Debug?': _IsDebugMode,
}


def SplitMeta(meta):
    """
    Split and validate metacharacters.

    Example: '{}' -> ('{', '}')

    This is public so the syntax highlighter and other tools can use it.
    """
    n = len(meta)
    if n % 2 == 1:
        raise ConfigurationError('%r has an odd number of metacharacters' % meta)
    return meta[:int(n/2)], meta[int(n/2):]


_token_re_cache = {}


def MakeTokenRegex(meta_left, meta_right):
    """
    Return a (compiled) regular expression for tokenization.

    Args:
      meta_left, meta_right: e.g. '{' and '}'

    - The regular expressions are memoized.
    - This function is public so the syntax highlighter can use it.
    """
    key = meta_left, meta_right
    if key not in _token_re_cache:
        # - Need () grouping for re.split
        # - For simplicity, we allow all characters except newlines inside
        #   metacharacters ({} / [])
        _token_re_cache[key] = re.compile(r'(' + re.escape(meta_left) + r'.+?' + re.escape(meta_right) + r')')
    return _token_re_cache[key]


# Examples:

(
    LITERAL_TOKEN,  # "Hi"
    SUBSTITUTION_TOKEN,  # {var|html}
    SECTION_TOKEN,  # {.section name}
    REPEATED_SECTION_TOKEN,  # {.repeated section name}
    PREDICATE_TOKEN,  # {.predicate?}
    ALTERNATES_TOKEN,  # {.or}
    OR_TOKEN,  # {.or}
    END_TOKEN,  # {.end}
) = list(range(8))


def _MatchDirective(token):
    """
    Helper function for matching certain directives.
    """

    if token.startswith('.'):
        token = token[1:]
    else:
        return None, None  # Must start with .

    if token == 'alternates with':
        return ALTERNATES_TOKEN, token

    if token.startswith('or'):
        if token.strip() == 'or':
            return OR_TOKEN, None
        else:
            pred_str = token[2:].strip()
            return OR_TOKEN, pred_str

    if token == 'end':
        return END_TOKEN, None

    match = _SECTION_RE.match(token)
    if match:
        repeated, section_name = match.groups()
        if repeated:
            return REPEATED_SECTION_TOKEN, section_name
        else:
            return SECTION_TOKEN, section_name

    # {.if plural?} and {.plural?} are synonyms.  The ".if" will read better for
    # expressions, for people who like that kind of dirty thing...
    if token.startswith('if '):
        return PREDICATE_TOKEN, token[3:].strip()
    if token.endswith('?'):
        return PREDICATE_TOKEN, token

    return None, None  # no match


def _Tokenize(template_str, meta_left, meta_right):
    """
    Yields tokens, which are 2-tuples (TOKEN_TYPE, token_string).
    """

    trimlen = len(meta_left)

    token_re = MakeTokenRegex(meta_left, meta_right)

    for line in template_str.splitlines(True):  # retain newlines
        tokens = token_re.split(line)

        # Check for a special case first.  If a comment or "block" directive is on a
        # line by itself (with only space surrounding it), then the space is
        # omitted.  For simplicity, we don't handle the case where we have 2
        # directives, say '{.end} # {#comment}' on a line.

        if len(tokens) == 3:
            # ''.isspace() == False, so work around that
            if (tokens[0].isspace() or not tokens[0]) and \
               (tokens[2].isspace() or not tokens[2]):
                token = tokens[1][trimlen:-trimlen]

                if token.startswith('#'):
                    continue  # The whole line is omitted

                token_type, token = _MatchDirective(token)
                if token_type is not None:
                    yield token_type, token  # Only yield the token, not space
                    continue

        # The line isn't special; process it normally.

        for i, token in enumerate(tokens):
            if i % 2 == 0:
                yield LITERAL_TOKEN, token

            else:  # It's a "directive" in metachracters

                assert token.startswith(meta_left), repr(token)
                assert token.endswith(meta_right), repr(token)
                token = token[trimlen:-trimlen]

                # It's a comment
                if token.startswith('#'):
                    continue

                if token.startswith('.'):

                    literal = {
                        '.meta-left': meta_left,
                        '.meta-right': meta_right,
                        '.space': ' ',
                        '.tab': '\t',
                        '.newline': '\n',
                    }.get(token)

                    if literal is not None:
                        yield LITERAL_TOKEN, literal
                        continue

                    token_type, token = _MatchDirective(token)
                    if token_type is not None:
                        yield token_type, token

                else:  # Now we know the directive is a substitution.
                    yield SUBSTITUTION_TOKEN, token


def CompileTemplate(template_str, builder=None, meta='{}', format_char='|', more_formatters=lambda x: None, more_predicates=lambda x: None, default_formatter='str'):
    """
    Compile the template string, calling methods on the 'program builder'.

    Args:
      template_str: The template string.  It should not have any compilation
          options in the header -- those are parsed by FromString/FromFile

      builder: The interface of _ProgramBuilder isn't fixed.  Use at your own
          risk.

      meta: The metacharacters to use, e.g. '{}', '[]'.

      more_formatters:
          Something that can map format strings to formatter functions.  One of:
            - A plain dictionary of names -> functions  e.g. {'html': html_escape}
            - A higher-order function which takes format strings and returns
              formatter functions.  Useful for when formatters have parsed
              arguments.
            - A FunctionRegistry instance for the most control.  This allows
              formatters which takes contexts as well.

      more_predicates:
          Like more_formatters, but for predicates.

      default_formatter: The formatter to use for substitutions that are missing a
          formatter.  The 'str' formatter the "default default" -- it just tries
          to convert the context value to a string in some unspecified manner.

    Returns:
      The compiled program (obtained from the builder)

    Raises:
      The various subclasses of CompilationError.  For example, if
      default_formatter=None, and a variable is missing a formatter, then
      MissingFormatter is raised.

    This function is public so it can be used by other tools, e.g. a syntax
    checking tool run before submitting a template to source control.
    """
    builder = builder or _ProgramBuilder(more_formatters, more_predicates)
    meta_left, meta_right = SplitMeta(meta)

    # : is meant to look like Python 3000 formatting {foo:.3f}.  According to
    # PEP 3101, that's also what .NET uses.
    # | is more readable, but, more importantly, reminiscent of pipes, which is
    # useful for multiple formatters, e.g. {name|js-string|html}
    if format_char not in (':', '|'):
        raise ConfigurationError('Only format characters : and | are accepted (got %r)' % format_char)

    # If we go to -1, then we got too many {end}.  If end at 1, then we're missing
    # an {end}.
    balance_counter = 0

    for token_type, token in _Tokenize(template_str, meta_left, meta_right):

        if token_type == LITERAL_TOKEN:
            if token:
                builder.Append(token)
            continue

        if token_type in (SECTION_TOKEN, REPEATED_SECTION_TOKEN):
            builder.NewSection(token_type, token)
            balance_counter += 1
            continue

        if token_type == PREDICATE_TOKEN:
            # Everything of the form {.predicate?} starts a new predicate section
            block_made = builder.NewPredicateSection(token)
            balance_counter += 1
            continue

        if token_type == OR_TOKEN:
            builder.NewOrClause(token)
            continue

        if token_type == ALTERNATES_TOKEN:
            builder.AlternatesWith()
            continue

        if token_type == END_TOKEN:
            balance_counter -= 1
            if balance_counter < 0:
                # TODO: Show some context for errors
                raise TemplateSyntaxError('Got too many %send%s statements.  You may have mistyped an '
                                          "earlier 'section' or 'repeated section' directive." % (meta_left, meta_right), )
            builder.EndSection()
            continue

        if token_type == SUBSTITUTION_TOKEN:
            parts = token.split(format_char)
            if len(parts) == 1:
                if default_formatter is None:
                    raise MissingFormatter('This template requires explicit formatters.')
                # If no formatter is specified, the default is the 'str' formatter,
                # which the user can define however they desire.
                name = token
                formatters = [default_formatter]
            else:
                name = parts[0]
                formatters = parts[1:]

            builder.AppendSubstitution(name, formatters)

    if balance_counter != 0:
        raise TemplateSyntaxError('Got too few %send%s statements' % (meta_left, meta_right))

    return builder.Root()


_OPTION_RE = re.compile(r'^([a-zA-Z\-]+):\s*(.*)')
# TODO: whitespace mode, etc.
_OPTION_NAMES = ['meta', 'format-char', 'default-formatter', 'undefined-str']


def FromString(s, more_formatters=lambda x: None, _constructor=None):
    """
    Like FromFile, but takes a string.
    """

    f = StringIO(s)
    return FromFile(f, more_formatters=more_formatters, _constructor=_constructor)


def FromFile(f, more_formatters=lambda x: None, _constructor=None):
    """
    Parse a template from a file, using a simple file format.

    This is useful when you want to include template options in a data file,
    rather than in the source code.

    The format is similar to HTTP or E-mail headers.  The first lines of the file
    can specify template options, such as the metacharacters to use.  One blank
    line must separate the options from the template body.

    Example:

      default-formatter: none
      meta: {{}}
      format-char: :
      <blank line required>
      Template goes here: {{variable:html}}

    Args:
      f: A file handle to read from.  Caller is responsible for opening and
      closing it.
    """
    _constructor = _constructor or Template

    options = {}

    # Parse lines until the first one that doesn't look like an option
    while True:
        line = f.readline()
        match = _OPTION_RE.match(line)
        if match:
            name, value = match.group(1), match.group(2)

            # Accept something like 'Default-Formatter: raw'.  This syntax is like
            # HTTP/E-mail headers.
            name = name.lower()

            if name in _OPTION_NAMES:
                name = name.replace('-', '_')
                value = value.strip()
                if name == 'default_formatter' and value.lower() == 'none':
                    value = None
                options[name] = value
            else:
                break
        else:
            break

    if options:
        if line.strip():
            raise CompilationError('Must be one blank line between template options and body (got %r)' % line)
        body = f.read()
    else:
        # There were no options, so no blank line is necessary.
        body = line + f.read()

    return _constructor(body, more_formatters=more_formatters, **options)


class Template(object):
    """
    Represents a compiled template.

    Like many template systems, the template string is compiled into a program,
    and then it can be expanded any number of times.  For example, in a web app,
    you can compile the templates once at server startup, and use the expand()
    method at request handling time.  expand() uses the compiled representation.

    There are various options for controlling parsing -- see CompileTemplate.
    Don't go crazy with metacharacters.  {}, [], {{}} or <> should cover nearly
    any circumstance, e.g. generating HTML, CSS XML, JavaScript, C programs, text
    files, etc.
    """
    def __init__(self, template_str, builder=None, undefined_str=None, **compile_options):
        """
        Args: template_str: The template string. undefined_str: A string to
        appear in the output when a variable to be substituted is missing.  If
        None, UndefinedVariable is raised. (Note: This is not really a
        compilation option, because affects template expansion rather than
        compilation.  Nonetheless we make it a constructor argument rather than
        an .expand() argument for simplicity.)

        It also accepts all the compile options that CompileTemplate
        does.
        """
        self._program = CompileTemplate(template_str, builder=builder, **compile_options)
        self.undefined_str = undefined_str

    #
    # Public API
    #

    def render(self, data_dict, callback):
        """
        Low level method to expands the template piece by piece.

        Args:
          data_dict: The JSON data dictionary.
          callback: A callback which should be called with each expanded token.

        Example: You can pass 'f.write' as the callback to write directly to a file
        handle.
        """
        context = _ScopedContext(data_dict, self.undefined_str)
        _Execute(self._program.Statements(), context, callback)

    def expand(self, *args, **kwargs):
        """
        Expands the template with the given data dictionary, returning a
        string.

        This is a small wrapper around render(), and is the most convenient
        interface.

        Args:
          The JSON data dictionary.  Like the builtin dict() constructor, it can
          take a single dictionary as a positional argument, or arbitrary keyword
          arguments.

        Returns:
          The return value could be a str() or unicode() instance, depending on the
          the type of the template string passed in, and what the types the strings
          in the dictionary are.
        """
        if args:
            if len(args) == 1:
                data_dict = args[0]
            else:
                raise TypeError('expand() only takes 1 positional argument (got %s)' % args)
        else:
            data_dict = kwargs

        tokens = []
        self.render(data_dict, tokens.append)
        return ''.join(tokens)

    def tokenstream(self, data_dict):
        """
        Yields a list of tokens resulting from expansion.

        This may be useful for WSGI apps.  NOTE: In the current implementation, the
        entire expanded template must be stored memory.

        NOTE: This is a generator, but JavaScript doesn't have generators.
        """
        tokens = []
        self.render(data_dict, tokens.append)
        for token in tokens:
            yield token


def _DoRepeatedSection(args, context, callback):
    """
    {repeated section foo}
    """

    block = args

    items = context.PushSection(block.section_name)
    # TODO: if 'items' is a dictionary, allow @name and @value.

    if items:
        if not isinstance(items, list):
            raise EvaluationError('Expected a list; got %s' % type(items))

        last_index = len(items) - 1
        statements = block.Statements()
        alt_statements = block.Statements('alternates with')
        try:
            i = 0
            while True:
                context.Next()
                # Execute the statements in the block for every item in the list.
                # Execute the alternate block on every iteration except the last.  Each
                # item could be an atom (string, integer, etc.) or a dictionary.
                _Execute(statements, context, callback)
                if i != last_index:
                    _Execute(alt_statements, context, callback)
                i += 1
        except StopIteration:
            pass

    else:
        _Execute(block.Statements('or'), context, callback)

    context.Pop()


def _DoSection(args, context, callback):
    """
    {section foo}
    """
    block = args
    # If a section present and "true", push the dictionary onto the stack as the
    # new context, and show it
    if context.PushSection(block.section_name):
        _Execute(block.Statements(), context, callback)
        context.Pop()
    else:  # missing or "false" -- show the {.or} section
        context.Pop()
        _Execute(block.Statements('or'), context, callback)


def _DoPredicates(args, context, callback):
    """
    {.predicate?}

    Here we execute the first clause that evaluates to true, and then
    stop.
    """
    block = args
    value = context.Lookup('@')
    for (predicate, args, func_type), statements in block.clauses:
        if func_type == ENHANCED_FUNC:
            do_clause = predicate(value, context, args)
        else:
            do_clause = predicate(value)

        if do_clause:
            _Execute(statements, context, callback)
            break


def _DoSubstitute(args, context, callback):
    """
    Variable substitution, e.g. {foo}
    """

    name, formatters = args

    # So we can have {.section is_new}new since {@}{.end}.  Hopefully this idiom
    # is OK.
    if name == '@':
        value = context.Lookup('@')
    else:
        try:
            value = context.Lookup(name)
        except TypeError as e:
            raise EvaluationError('Error evaluating %r in context %r: %r' % (name, context, e))

    for func, args, func_type in formatters:
        try:
            if func_type == ENHANCED_FUNC:
                value = func(value, context, args)
            else:
                value = func(value)
        except KeyboardInterrupt:
            raise
        except Exception as e:
            raise EvaluationError('Formatting value %r with formatter %s raised exception: %r' % (value, formatters, e), original_exception=e)

    # TODO: Require a string/unicode instance here?
    if value is None:
        raise EvaluationError('Evaluating %r gave None value' % name)
    callback(value)


def _Execute(statements, context, callback):
    """
    Execute a bunch of template statements in a ScopedContext.

    Args:
      callback: Strings are "written" to this callback function.

    This is called in a mutually recursive fashion.
    """

    for i, statement in enumerate(statements):
        if isinstance(statement, six.string_types):
            callback(statement)
        else:
            # In the case of a substitution, args is a pair (name, formatters).
            # In the case of a section, it's a _Section instance.
            try:
                func, args = statement
                func(args, context, callback)
            except UndefinedVariable as e:
                # Show context for statements
                start = max(0, i - 3)
                end = i + 3
                e.near = statements[start:end]
                raise


def expand(template_str, dictionary, **kwargs):
    """
    Free function to expands a template string with a data dictionary.

    This is useful for cases where you don't care about saving the
    result of compilation (similar to re.match('.*', s) vs
    DOT_STAR.match(s))
    """
    t = Template(template_str, **kwargs)
    return t.expand(dictionary)



import re
import json
import time
import six

from six import PY2, b

from functools import wraps

from twisted.web.resource import Resource
from twisted.web.server import NOT_DONE_YET
from twisted.internet.defer import Deferred
from twisted.python import log as twlog
from twisted.python.failure import Failure

_Debug = False


def _to_text(v):
    if isinstance(v, six.binary_type):
        v = v.decode()
    if not isinstance(v, six.text_type):
        v = six.text_type(v)
    return v


def _to_json(output_object):
    return (json.dumps(
        output_object,
        indent=2,
        separators=(',', ': '),
        sort_keys=True,
        default=_to_text,
    ) + '\n').encode()


class _JsonResource(Resource):
    _result = ''
    isLeaf = True

    def __init__(self, result, executed, *args, **kwargs):
        Resource.__init__(self, *args, **kwargs)
        self._result = result
        self._executed = executed

    def _setHeaders(self, request):
        """
        Those headers will allow you to call API methods from web browsers, they require CORS:
            https://en.wikipedia.org/wiki/Cross-origin_resource_sharing
        """
        request.responseHeaders.addRawHeader(b('content-type'), b('application/json'))
        allow_origin = 'localhost'
        if _Debug:
            allow_origin = 'http://localhost:8080'
        if True:
            # TODO: !!! Find a solution here to protect API from non-bitdust UI apps !!!
            allow_origin = '*'
        request.responseHeaders.addRawHeader(b('Access-Control-Allow-Origin'), b(allow_origin))
        request.responseHeaders.addRawHeader(b('Access-Control-Allow-Methods'), b('GET, POST, PUT, DELETE'))
        request.responseHeaders.addRawHeader(b('Access-Control-Allow-Headers'), b('x-prototype-version,x-requested-with'))
        request.responseHeaders.addRawHeader(b('Access-Control-Max-Age'), b('2520'))
        return request

    def render(self, request):
        self._setHeaders(request)
        # this will just add one extra field to the response to populate how fast that API call was processed
        self._result['execution'] = '%3.6f' % (time.time() - self._executed)
        return _to_json(self._result)


class _DelayedJsonResource(_JsonResource):
    """
    If your API method returned `Deferred` object instead of final result
    we can wait for the result and then return it in API response.
    """
    def __init__(self, result, executed, result_defer, *args, **kwargs):
        _JsonResource.__init__(self, result, executed, *args, **kwargs)
        self._result_defer = result_defer

    def _cb(self, result, request):
        if _Debug:
            twlog.msg('txrestapi callback with json result: %r' % result)
        self._setHeaders(request)
        result['execution'] = '%3.6f' % (time.time() - self._executed)
        raw = _to_json(result)
        if not request.channel:
            if _Debug:
                twlog.err('REST API connection channel already closed')
            return None
        request.write(raw)
        request.finish()
        return result

    def _eb(self, err, request):
        if _Debug:
            twlog.err('txrestapi error : %r' % err)
        self._setHeaders(request)
        execution = '%3.6f' % (time.time() - self._executed)
        err_msg = err.getErrorMessage() if isinstance(err, Failure) else str(err)
        raw = _to_json(dict(
            status='ERROR',
            execution=execution,
            errors=[
                err_msg,
            ],
        ))
        if not request.channel:
            if _Debug:
                twlog.err('REST API connection channel already closed')
            return None
        request.write(raw)
        request.finish()
        return None

    def render(self, request):
        self._result_defer.addCallback(self._cb, request)
        self._result_defer.addErrback(self._eb, request)
        return NOT_DONE_YET


def maybeResource(f):
    @wraps(f)
    def inner(*args, **kwargs):
        _executed = time.time()
        try:
            result = f(*args, **kwargs)

        except Exception as exc:
            return _JsonResource(
                result=dict(
                    status='ERROR',
                    errors=[
                        str(exc),
                    ],
                ),
                executed=_executed,
            )

        if isinstance(result, Deferred):
            return _DelayedJsonResource(
                result=None,
                executed=_executed,
                result_defer=result,
            )

        if not isinstance(result, Resource):
            result = _JsonResource(
                result=result,
                executed=_executed,
            )

        return result

    return inner


class JsonAPIResource(Resource):

    _registry = None

    def __new__(cls, *args, **kwds):
        instance = super().__new__(cls, *args, **kwds)
        instance._registry = []
        for name in dir(instance):
            attribute = getattr(instance, name)
            annotations = getattr(attribute, '__txrestapi__', [])
            for annotation in annotations:
                method, regex = annotation
                instance.register(method, regex, attribute)
        return instance

    def __init__(self, *args, **kwargs):
        if PY2:
            self._registry = []
        Resource.__init__(self, *args, **kwargs)

    def _get_callback(self, request):
        path_to_check = getattr(request, '_remaining_path', request.path)
        if not isinstance(path_to_check, six.text_type):
            path_to_check = path_to_check.decode()
        for m, r, cb in self._registry:
            if m == request.method or m == b('ALL'):
                result = r.search(path_to_check)
                if result:
                    request._remaining_path = path_to_check[result.span()[1]:]
                    return cb, result.groupdict()
        return None, None

    def log_request(self, request, callback, args):
        return None

    def register(self, method, regex, callback):
        if not isinstance(regex, six.text_type):
            regex = regex.decode()
        self._registry.append((method, re.compile(regex), callback))

    def unregister(self, method=None, regex=None, callback=None):
        if regex is not None:
            if not isinstance(regex, six.text_type):
                regex = regex.decode()
            regex = re.compile(regex)
        for m, r, cb in self._registry[:]:
            if not method or (method and m == method):
                if not regex or (regex and r == regex):
                    if not callback or (callback and cb == callback):
                        self._registry.remove((m, r, cb))

    def getChild(self, name, request):
        r = self.children.get(name, None)
        if r is None:
            # Go into the thing
            callback, args = self._get_callback(request)
            self.log_request(request, callback, args)
            if callback is None:
                return _JsonResource(
                    dict(
                        status='ERROR',
                        errors=[
                            'path %r not found' % name,
                        ],
                    ),
                    time.time(),
                )
            else:
                return maybeResource(callback)(request, **args)
        else:
            return r
